import copy
from typing import Any
from dataclasses import dataclass

import numpy as np

from mirp.settings.feature_parameters import FeatureExtractionSettingsClass
from mirp.settings.utilities import setting_def


@dataclass
class ImageTransformationSettingsClass:
    """
    Parameters related to image transformation using filters. Many parameters are conditional on the selected image
    filter (``filter_kernels``). Filters and their parameters are defined `here <https://arxiv.org/abs/2006.05470>`_
    By default, only statistical features are computed from filtered images.

    .. note::
        Many feature extraction parameters are copied from
        :class:`~mirp.settings.feature_parameters.FeatureExtractionSettingsClass`, except
        ``response_map_feature_families``, ``response_map_discretisation_method`` and
        ``response_map_discretisation_n_bins``. If other parameters need to be changed from their default settings,
        first create an object of the current class (
        :class:`~mirp.settings.transformation_parameters.ImageTransformationSettingsClass`), and then update the
        attributes.

    Parameters
    ----------
    by_slice: bool, optional, default: False
        Defines whether calculations should be performed in 2D (True) or 3D (False).
        See :class:`~mirp.settings.general_parameters.GeneralSettingsClass`.

    ibsi_compliant: bool, optional, default: True
        Limits use of filters to those that exist in the IBSI reference standard.

    response_map_feature_families: str or list of str, optional, default: "statistics"
        Determines the feature families for which features are computed from response maps (filtered images). Radiomics
        features are implemented as defined in the IBSI reference manual. The following feature families can be
        computed from response maps:

        * Local intensity features: "li", "loc.int", "loc_int", "local_int", and "local_intensity".
        * Intensity-based statistical features: "st", "stat", "stats", "statistics", and "statistical".
        * Intensity histogram features: "ih", "int_hist", "int_histogram", and "intensity_histogram".
        * Intensity-volume histogram features: "ivh", "int_vol_hist", and "intensity_volume_histogram".
        * Grey level co-occurrence matrix (GLCM) features: "cm", "glcm", "grey_level_cooccurrence_matrix",
          and "cooccurrence_matrix".
        * Grey level run length matrix (GLRLM) features: "rlm", "glrlm", "grey_level_run_length_matrix", and
          "run_length_matrix".
        * Grey level size zone matrix (GLSZM) features: "szm", "glszm", "grey_level_size_zone_matrix", and
          "size_zone_matrix".
        * Grey level distance zone matrix (GLDZM) features: "dzm", "gldzm", "grey_level_distance_zone_matrix", and
          "distance_zone_matrix".
        * Neighbourhood grey tone difference matrix (NGTDM) features: "tdm", "ngtdm",
          "neighbourhood_grey_tone_difference_matrix", and "grey_tone_difference_matrix".
        * Neighbouring grey level dependence matrix (NGLDM) features: "ldm", "ngldm",
          "neighbouring_grey_level_dependence_matrix", and "grey_level_dependence_matrix".

        In addition, the following tags can be used:

        * "none": no features are computed.
        * "all": all features are computed.

        A list of tags may be provided to select multiple feature families. Morphological features are not computed
        from response maps (filtered images), because these are mask-based and are invariant to filtering.

    response_map_discretisation_method: {"fixed_bin_number", "fixed_bin_size", "fixed_bin_size_pyradiomics", "none"}, optional, default: "fixed_bin_number"
        Method used for discretising intensities. Used to compute intensity histogram as well as texture features.
        The setting is ignored if none of these feature families are being computed. The following options are
        available:

        * "fixed_bin_number": The intensity range within the mask is divided into a fixed number of bins,
          defined by the ``base_discretisation_bin_width`` parameter.
        * "fixed_bin_size": The intensity range is divided into bins with a fixed width, defined using the
          ``base_discretisation_bin_width`` parameter. The lower bound of the range is determined from the lower
          bound of the mask resegmentation range, see the ``resegmentation_intensity_range`` in
          :class:`~mirp.settings.resegmentation_parameters.ResegmentationSettingsClass`. Other images,
          including MRI, normalised CT and PET images and filtered images, do not have a default value, and bins are
          created from using the minimum intensity as lower bound.
        * "fixed_bin_size_pyradiomics": The intensity range is divided into bins with a fixed width. This follows the
          non-IBSI compliant implementation in the pyradiomics package.
        * "none": The intensity range is not discretised into bins. This method can only be used if the image
          intensities are integer and strictly positive.

        Multiple discretisation methods can be specified as a list to yield features according to each method.

        .. note::
            Use of the "fixed_bin_size", "fixed_bin_size_pyradiomics", and "none" discretisation methods is discouraged
            for transformed images. Due to transformation, a direct link to any meaningful quantity represented by the
            intensity of the original image (e.g. Hounsfield Units for CT, Standardised Uptake Value for PET) is lost.

    response_map_discretisation_n_bins: int or list of int, optional, default: 16
        Number of bins used for the "fixed_bin_number" discretisation method. Multiple values can be specified in a
        list to yield features according to each number of bins.

    response_map_discretisation_bin_width: float or list of float, optional
        Width of each bin in the "fixed_bin_size" and "fixed_bin_size_pyradiomics" discretisation methods. Multiple
        values can be specified in a list to yield features according to each bin width.

    filter_kernels: str or list of str, optional, default: None
        Names of the filters applied to the original image to create response maps (filtered images). Filter
        implementation follows the IBSI reference manual. The following filters are supported:

        * Mean filters: "mean"
        * Gaussian filters: "gaussian", "riesz_gaussian", and "riesz_steered_gaussian"
        * Laplacian-of-Gaussian filters: "laplacian_of_gaussian", "log", "riesz_laplacian_of_gaussian",
          "riesz_log", "riesz_steered_laplacian_of_gaussian", and "riesz_steered_log".
        * Laws kernels: "laws"
        * Gabor kernels: "gabor", "riesz_gabor", and "riesz_steered_gabor"
        * Separable wavelets: "separable_wavelet"
        * Non-separable wavelets: "nonseparable_wavelet", "riesz_nonseparable_wavelet",
          and "riesz_steered_nonseparable_wavelet"
        * Local binary patterns: "lbp", "lbp_2d", "lbp_3d"
        * Function transformations: "pyradiomics_square", "pyradiomics_square_root",
          and "pyradiomics_logarithm", "pyradiomics_exponential"

        Filters with names that preceded by "riesz" undergo a Riesz transformation. If the filter name is preceded by
        "riesz_steered", a steerable riesz filter is used.

        More than one filter name can be provided. By default, no filters are selected, and image transformation is
        skipped.

        .. note::
            There is no IBSI reference standard for Gaussian filters. However, the filter implementation is relatively
            straightforward, and most likely reproducible.

        .. warning::
            Riesz transformation and steerable riesz transformations are experimental. The implementation of these
            filter transformations is complex. Since there is no corresponding IBSI reference standard, any feature
            derived from response maps (filtered images) of Riesz transformations is unlikely to be reproducible.

        .. warning::
            Local binary patterns are not part of the IBSI reference standard. Set `ibsi_compliant = False` to use
            this filter.

        .. warning::
            Function transformations (square, square root, logarithm, exponential) do not have an IBSI reference
            standard. These transformations follow the definition in pyradiomics, and have been implemented for
            validation purposes.

    boundary_condition: {"reflect", "constant", "nearest", "mirror", "wrap"}, optional, default: "mirror"
        Sets the boundary condition, which determines how filters behave at the edge of an image. MIRP uses
        the same nomenclature for boundary conditions as scipy.ndimage. See the ``mode`` parameter of
        `scipy.ndimage.convolve
        <https://docs.scipy.org/doc/scipy/reference/generated/scipy.ndimage.convolve.html#scipy.ndimage.convolve>`_

    separable_wavelet_families: str or list str
        Name of separable wavelet kernel as implemented in the ``pywavelets`` package. See `pywt.wavelist(
        kind="discrete") <https://pywavelets.readthedocs.io/en/latest/ref/wavelets.html#built-in-wavelets-wavelist>`_
        for options.

    separable_wavelet_set: str or list of str, optional
        Filter orientation of separable wavelets. Allows for specifying combinations for high and low-pass filters.
        For 2D (``by_slice=True``) filters, the following sets are possible: "hh", "hl", "lh", "ll" (y-x directions).
        For 3D (``by_slice=False``) filters, the set of possibilities is larger: "hhh", "hhl", "hlh", "lhh", "hll",
        "lhl", "llh", "lll". More than one  orientation may be set. Default: "hh" (2d) or "hhh (3d).

    separable_wavelet_stationary: bool, optional, default: True
        Determines if wavelets are stationary or not. Stationary wavelets maintain the image dimensions after
        decomposition.

    separable_wavelet_decomposition_level: int or list of int, optional, default: 1
        Sets the wavelet decomposition level. For the first decomposition level, the base image is used as input to
        generate a  response map (filtered image). For decomposition levels greater than 1, the low-pass image from the
        previous level is used as input. More than 1 value may be specified in a list.

    separable_wavelet_rotation_invariance: bool, optional, default: True
        Determines whether separable filters are applied in a pseudo-rotational invariant manner. This generates
        permutations of the filter and, as a consequence, additional response maps (filtered images). These maps are
        then merged using the pooling method (``separable_wavelet_pooling_method``).

    separable_wavelet_pooling_method: {"max", "min", "mean", "sum"}, optional, default: "max"
        Response maps are pooled to create a rotationally invariant response map (filtered image). This sets the
        method for pooling.

        * "max": Each voxel of the pooled response map represents the maximum value for that voxel in the underlying
          response maps.
        * "min": Each voxel of the pooled response map represents the minimum value for that voxel in the underlying
          response maps.
        * "mean": Each voxel of the pooled response map represents the mean value for that voxel in the underlying
          response maps. For band-pass and high-pass filters, this will likely result in values close to 0.0,
          and "max" or "min" pooling methods should be used instead.
        * "sum": Each voxel of the pooled response map is the sum of intensities for that voxel in the underlying
          response maps. Similar to the "mean" pooling method, but without the normalisation.

    separable_wavelet_boundary_condition: str, optional, default: "mirror"
        Sets the boundary condition for separable wavelets. This supersedes any value set by the general
        ``boundary_condition`` parameter. See the ``boundary_condition`` parameter above for all valid options.

    nonseparable_wavelet_families: {"shannon", "simoncelli"}
        Name of non-separable wavelet kernels used for image transformation. Shannon and Simoncelli wavelets are
        implemented.

    nonseparable_wavelet_decomposition_level: int or list of int, optional, default: 1
        Sets the wavelet decomposition level. Unlike the decomposition level in separable wavelets, decomposition of
        non-separable wavelets is purely a filter-based operation.

    nonseparable_wavelet_response: {"modulus", "abs", "magnitude", "angle", "phase", "argument", "real", "imaginary"}, optional, default: "real"
        Nonseparable wavelets produce response maps with complex numbers. The complex-valued response map is
        converted to a real-valued response map (filtered image) using the specified method. "modulus", "abs",
        "magnitude" are synonymous, as are "angle", "phase", and "argument". "real" selects the real component of the
        complex values, and "imaginary" selects the imaginary component.

    nonseparable_wavelet_boundary_condition: str, optional, default: "mirror"
        Sets the boundary condition for non-separable wavelets. This supersedes any value set by the general
        ``boundary_condition`` parameter. See the ``boundary_condition`` parameter above for all valid options.

    gaussian_sigma: float or list of float, optional
         Width of the Gaussian filter in physical dimensions (e.g. mm). Multiple values can be specified.

    gaussian_kernel_truncate: float, optional, default: 4.0
        Width, in units of sigma, at which the filter is truncated.

    gaussian_kernel_boundary_condition: str, optional, default: "mirror"
        Sets the boundary condition for Gaussian filters. This supersedes any value set by the general
        ``boundary_condition`` parameter. See the ``boundary_condition`` parameter above for all valid options.

    laplacian_of_gaussian_sigma: float or list of float, optional
        Width of the Gaussian filter in physical dimensions (e.g. mm). Multiple values can be specified.

    laplacian_of_gaussian_kernel_truncate: float, optional, default: 4.0
        Width, in sigma, at which the filter is truncated.

    laplacian_of_gaussian_pooling_method: {"max", "min", "mean", "sum", "none"}, optional, default: "none"
        Determines whether and how response maps (filtered images) for filters with different widths (
        ``laplacian_of_gaussian_sigma``) are pooled.

        * "max": Each voxel of the pooled response map represents the maximum value for that voxel in the underlying
          response maps.
        * "min": Each voxel of the pooled response map represents the minimum value for that voxel in the underlying
          response maps.
        * "mean": Each voxel of the pooled response map represents the mean value for that voxel in the underlying
          response maps. For band-pass and high-pass filters, this will likely result in values close to 0.0,
          and "max" or "min" pooling methods should be used instead.
        * "sum": Each voxel of the pooled response map is the sum of intensities for that voxel in the underlying
          response maps. Similar to the "mean" pooling method, but without the normalisation.
        * "none": Each Laplacian-of-Gaussian response map is treated separately, without pooling.

    laplacian_of_gaussian_boundary_condition: str, optional, default: "mirror"
        Sets the boundary condition for Laplacian-of-Gaussian filters. This supersedes any value set by the general
        ``boundary_condition`` parameter. See the ``boundary_condition`` parameter above for all valid options.

    laws_kernel: str or list of str, optional
        Compute specific Laws kernels these typically are specific combinations of kernels such as L5S5E5,
        E5E5E5. The following kernels are available: 'l5', 'e5', 's5', 'w5', 'r5', 'l3', 'e3', 's3'. A combination of
        two kernels is expected for 2D (``by_slice=True``), whereas a kernel triplet is expected for 3D filters (
        ``by_slice=False``).

    laws_compute_energy: bool, optional, default: True
        Determine whether an energy image should be computed, or just the response map (filtered) image.

    laws_delta: int or list of int, optional, default: 7
        Delta for chebyshev distance between center voxel and neighbourhood boundary used to calculate energy maps.

    laws_rotation_invariance: bool, optional, default: True
        Determines whether separable filters are applied in a pseudo-rotational invariant manner. This generates
        permutations of the filter and, as a consequence, additional response maps (filtered images). These maps are
        then merged using the pooling method (``laws_pooling_method``).

    laws_pooling_method:  {"max", "min", "mean", "sum"}, optional, default: "max"
        Response maps are pooled to create a rotationally invariant response map (filtered image). This sets the
        method for pooling.

        * "max": Each voxel of the pooled response map represents the maximum value for that voxel in the underlying
          response maps.
        * "min": Each voxel of the pooled response map represents the minimum value for that voxel in the underlying
          response maps.
        * "mean": Each voxel of the pooled response map represents the mean value for that voxel in the underlying
          response maps. For band-pass and high-pass filters, this will likely result in values close to 0.0,
          and "max" or "min" pooling methods should be used instead.
        * "sum": Each voxel of the pooled response map is the sum of intensities for that voxel in the underlying
          response maps. Similar to the "mean" pooling method, but without the normalisation.

    laws_boundary_condition: str, optional, default: "mirror"
        Sets the boundary condition for Laws filters. This supersedes any value set by the general
        ``boundary_condition`` parameter. See the ``boundary_condition`` parameter above for all valid options.

    gabor_sigma: float or list of float, optional
        Width of the Gaussian envelope in physical dimensions (e.g. mm). Multiple values can be specified.

    gabor_lambda: float or list of float, optional
        Wavelength of the oscillator component of the Gabor filter, in physical dimensions (e.g. mm).

    gabor_gamma: float or list of float, optional, default: 1.0
        Eccentricity parameter of the Gaussian envelope of the Gabor kernel. Defines width of y-axis relative to
        x-axis for 0-angle Gabor kernel. Default: 1.0

    gabor_theta: float or list of flaot, optional, default: 0.0
        Initial angle of the Gabor filter in degrees (not radians). Multiple angles can be provided.

    gabor_theta_step: float, optional, default: None
        Angle step size in degrees for in-plane rotational invariance. A value of 0.0 or None (default) disables
        stepping.

    gabor_response: {"modulus", "abs", "magnitude", "angle", "phase", "argument", "real", "imaginary"}, optional, default: "modulus"
        Type of response map (filtered image) created by Gabor filters. Gabor kernels consist of complex numbers,
        and the response map will be complex as well. The complex-valued response map is converted to a real-valued
        response map using the specified method.

    gabor_rotation_invariance: bool, optional, default: False
        Determines whether (2D) Gabor filters are applied in a pseudo-rotational invariant manner. If True,
        Gabor filters are applied in each of the orthogonal planes.

    gabor_pooling_method: {"max", "min", "mean", "sum"}, optional, default: "max"
        Response maps are pooled to create a rotationally invariant response map (filtered image). This sets the
        method for pooling.

        * "max": Each voxel of the pooled response map represents the maximum value for that voxel in the underlying
          response maps.
        * "min": Each voxel of the pooled response map represents the minimum value for that voxel in the underlying
          response maps.
        * "mean": Each voxel of the pooled response map represents the mean value for that voxel in the underlying
          response maps. For band-pass and high-pass filters, this will likely result in values close to 0.0,
          and "max" or "min" pooling methods should be used instead.
        * "sum": Each voxel of the pooled response map is the sum of intensities for that voxel in the underlying
          response maps. Similar to the "mean" pooling method, but without the normalisation.

    gabor_boundary_condition: str, optional, default: "mirror"
        Sets the boundary condition for Gabor filters. This supersedes any value set by the general
        ``boundary_condition`` parameter. See the ``boundary_condition`` parameter above for all valid options.

    mean_filter_kernel_size: int or list of int, optional
        Length of the kernel in pixels. Multiple values can be specified to create multiple response maps (filtered
        images).

    mean_filter_boundary_condition: str, optional, default: "mirror"
        Sets the boundary condition for mean filters. This supersedes any value set by the general
        ``boundary_condition`` parameter. See the ``boundary_condition`` parameter above for all valid options.

    riesz_filter_order: float, list of float or list of list of float, optional
        Riesz-transformation order. If required, should be a 2 (2D filter), or 3-element (3D filter) integer
        vector, e.g. [0,0,1]. Multiple sets can be provided by nesting the list, e.g. [[0, 0, 1],
        [0, 1, 0]]. If an integer is provided, a set of filters is created. For example when
        riesz_filter_order = 2 and a 2D filter is used, the following Riesz-transformations are performed: [2,
        0], [1, 1] and [0, 2].

        .. note::
            Riesz filter order uses the numpy coordinate ordering and represents (z, y, x) directions.

    riesz_filter_tensor_sigma: float or list of float, optional
        Determines width of Gaussian filter used with Riesz filter banks.

    lbp_method: str or list of str, optional, default: "default"
        Method for computing local binary pattern filter. The following methods are supported:

        * "default": the default method for LBP, which directionally encodes patterns.
        * "rotation_invariant": computes the unique minimal encoding pattern by rotating over the ring. This method
          is rotationally invariant.
        * "variance": variance of patterns, similar to the ``"var"`` method in ``scikit-image``. This method is
          rotationally invariant.

        LBP are computed with nearest neighbourhood interpolation.

    lbp_filter_distance: float or list of float, optional, default: 1.0
        Euclidean distance for the local binary pattern filter, in voxel spacing. With a distance of 1.0,
        all neighbouring voxels are selected.

    **kwargs: dict, optional
        Unused keyword arguments.

    """

    def __init__(
            self,
            by_slice: bool,
            ibsi_compliant: bool = True,
            response_map_feature_settings: FeatureExtractionSettingsClass | None = None,
            response_map_feature_families: None | str | list[str] = "statistical",
            response_map_discretisation_method: None | str | list[str] = "fixed_bin_number",
            response_map_discretisation_n_bins: None | int | list[int] = 16,
            response_map_discretisation_bin_width: None | int | list[float] = None,
            filter_kernels: None | str | list[str] = None,
            boundary_condition: None | str = "mirror",
            separable_wavelet_families: None | str | list[str] = None,
            separable_wavelet_set: None | str | list[str] = None,
            separable_wavelet_stationary: bool = True,
            separable_wavelet_decomposition_level: None | int | list[int] = 1,
            separable_wavelet_rotation_invariance: bool = True,
            separable_wavelet_pooling_method: str = "max",
            separable_wavelet_boundary_condition: None | str = None,
            nonseparable_wavelet_families: None | str | list[str] = None,
            nonseparable_wavelet_decomposition_level: None | int | list[int] = 1,
            nonseparable_wavelet_response: None | str = "real",
            nonseparable_wavelet_boundary_condition: None | str = None,
            gaussian_sigma: None | float | list[float] = None,
            gaussian_kernel_truncate: None | float = 4.0,
            gaussian_kernel_boundary_condition: None | str = None,
            laplacian_of_gaussian_sigma: None | float | list[float] = None,
            laplacian_of_gaussian_kernel_truncate: None | float = 4.0,
            laplacian_of_gaussian_pooling_method: str = "none",
            laplacian_of_gaussian_boundary_condition: None | str = None,
            laws_kernel: None | str | list[str] = None,
            laws_delta: int | list[int] = 7,
            laws_compute_energy: bool = True,
            laws_rotation_invariance: bool = True,
            laws_pooling_method: str = "max",
            laws_boundary_condition: None | str = None,
            gabor_sigma: None | float | list[float] = None,
            gabor_lambda: None | float | list[float] = None,
            gabor_gamma: None | float | list[float] = 1.0,
            gabor_theta: None | float | list[float] = 0.0,
            gabor_theta_step: None | float = None,
            gabor_response: str = "modulus",
            gabor_rotation_invariance: bool = False,
            gabor_pooling_method: str = "max",
            gabor_boundary_condition: None | str = None,
            mean_filter_kernel_size: None | int | list[int] = None,
            mean_filter_boundary_condition: None | str = None,
            riesz_filter_order: None | int | list[int] = None,
            riesz_filter_tensor_sigma: None | float | list[float] = None,
            lbp_method: None | str | list[str] = "default",
            lbp_filter_distance: None | float | list[float] = 1.0,
            **kwargs
    ):
        # Set by slice
        self.by_slice: bool = by_slice

        # Set IBSI-compliance flag.
        self.ibsi_compliant: bool = ibsi_compliant

        # Check filter kernels
        if not isinstance(filter_kernels, list):
            filter_kernels = [filter_kernels]

        if any(filter_kernel is None for filter_kernel in filter_kernels):
            filter_kernels = None

        if filter_kernels is not None:
            # Check validity of the filter kernel names.
            valid_kernels: list[bool] = [ii in self.get_available_image_filters() for ii in filter_kernels]

            if not all(valid_kernels):
                raise ValueError(
                    f"One or more kernels are not implemented, or were spelled incorrectly: "
                    f"{', '.join([filter_kernel for ii, filter_kernel in enumerate(filter_kernels) if not valid_kernels[ii]])}")

        self.spatial_filters: None | list[str] = filter_kernels

        # Check families.
        if response_map_feature_families is None:
            response_map_feature_families = "none"

        if not isinstance(response_map_feature_families, list):
            response_map_feature_families = [response_map_feature_families]

        # Check which entries are valid.
        valid_families: list[bool] = [ii in [
            "li", "loc.int", "loc_int", "local_int", "local_intensity", "st", "stat", "stats", "statistics",
            "statistical", "ih", "int_hist", "int_histogram", "intensity_histogram",
            "ivh", "int_vol_hist", "intensity_volume_histogram", "cm", "glcm", "grey_level_cooccurrence_matrix",
            "cooccurrence_matrix", "rlm", "glrlm", "grey_level_run_length_matrix", "run_length_matrix",
            "szm", "glszm", "grey_level_size_zone_matrix", "size_zone_matrix", "dzm", "gldzm",
            "grey_level_distance_zone_matrix", "distance_zone_matrix", "tdm", "ngtdm",
            "neighbourhood_grey_tone_difference_matrix", "grey_tone_difference_matrix", "ldm", "ngldm",
            "neighbouring_grey_level_dependence_matrix", "grey_level_dependence_matrix", "all", "none"
        ] for ii in response_map_feature_families]

        if not all(valid_families):
            raise ValueError(
                f"One or more families in the response_map_feature_families parameter were not recognised: "
                f"{', '.join([response_map_feature_families[ii] for ii, is_valid in enumerate(valid_families) if not is_valid])}"
            )

        # Create a temporary feature settings object. If response_map_feature_settings is not present, this object is
        # used. Otherwise, response_map_feature_settings is copied, and then updated.
        if response_map_feature_settings is None:

            kwargs = copy.deepcopy(kwargs)
            kwargs.update({
                "base_feature_families": response_map_feature_families,
                "base_discretisation_method": response_map_discretisation_method,
                "base_discretisation_bin_width": response_map_discretisation_bin_width,
                "base_discretisation_n_bins": response_map_discretisation_n_bins
            })

            response_map_feature_settings = FeatureExtractionSettingsClass(
                by_slice=by_slice,
                no_approximation=False,
                ibsi_compliant=ibsi_compliant,
                **kwargs
            )

        # Set feature settings.
        self.feature_settings: FeatureExtractionSettingsClass = response_map_feature_settings

        # If "all" appears in feature_settings.families: exclude morphological features.
        if "all" in self.feature_settings.families:
            self.feature_settings.families = [
                "local_intensity", "statistics", "intensity_histogram", "intensity_volume_histogram", "glcm", "glrlm",
                "glszm", "gldzm", "ngtdm", "ngldm"
            ]

        # Check boundary condition.
        self.boundary_condition = boundary_condition
        self.boundary_condition: str = self.check_boundary_condition(
            boundary_condition,
            "boundary_condition")

        # Check mean filter settings
        if self.has_mean_filter():
            # Check filter size.
            if not isinstance(mean_filter_kernel_size, list):
                mean_filter_kernel_size = [mean_filter_kernel_size]

            if not all(isinstance(kernel_size, int) for kernel_size in mean_filter_kernel_size):
                raise TypeError(
                    f"All kernel sizes for the mean filter are expected to be integer values equal or "
                    f"greater than 1. Found: one or more kernel sizes that were not integers.")

            if not all(kernel_size >= 1 for kernel_size in mean_filter_kernel_size):
                raise ValueError(
                    f"All kernel sizes for the mean filter are expected to be integer values equal or "
                    f"greater than 1. Found: one or more kernel sizes less then 1.")

            # Check boundary condition
            mean_filter_boundary_condition = self.check_boundary_condition(
                mean_filter_boundary_condition,
                "mean_filter_boundary_condition")

        else:
            mean_filter_kernel_size = None
            mean_filter_boundary_condition = None

        self.mean_filter_size: None | list[int] = mean_filter_kernel_size
        self.mean_filter_boundary_condition: None | str = mean_filter_boundary_condition

        # Check Gaussian kernel settings.
        if self.has_gaussian_filter():
            # Check sigma.
            gaussian_sigma = self.check_sigma(
                gaussian_sigma,
                "gaussian_sigma")

            # Check filter truncation.
            gaussian_kernel_truncate = self.check_truncation(
                gaussian_kernel_truncate,
                "gaussian_kernel_truncate")

            # Check boundary condition
            gaussian_kernel_boundary_condition = self.check_boundary_condition(
                gaussian_kernel_boundary_condition,
                "gaussian_kernel_boundary_condition")

        else:
            gaussian_sigma = None
            gaussian_kernel_truncate = None
            gaussian_kernel_boundary_condition = None

        self.gaussian_sigma: None | list[float] = gaussian_sigma
        self.gaussian_sigma_truncate: None | float = gaussian_kernel_truncate
        self.gaussian_boundary_condition: None | str = gaussian_kernel_boundary_condition

        # Check laplacian-of-gaussian filter settings
        if self.has_laplacian_of_gaussian_filter():
            # Check sigma.
            laplacian_of_gaussian_sigma = self.check_sigma(
                laplacian_of_gaussian_sigma,
                "laplacian_of_gaussian_sigma")

            # Check filter truncation.
            laplacian_of_gaussian_kernel_truncate = self.check_truncation(
                laplacian_of_gaussian_kernel_truncate,
                "laplacian_of_gaussian_kernel_truncate")

            # Check pooling method.
            laplacian_of_gaussian_pooling_method = self.check_pooling_method(
                laplacian_of_gaussian_pooling_method,
                "laplacian_of_gaussian_pooling_method",
                allow_none=True)

            # Check boundary condition.
            laplacian_of_gaussian_boundary_condition = self.check_boundary_condition(
                laplacian_of_gaussian_boundary_condition, "laplacian_of_gaussian_boundary_condition")

        else:
            laplacian_of_gaussian_sigma = None
            laplacian_of_gaussian_kernel_truncate = None
            laplacian_of_gaussian_pooling_method = None
            laplacian_of_gaussian_boundary_condition = None

        self.log_sigma: None | list[float] = laplacian_of_gaussian_sigma
        self.log_sigma_truncate: None | float = laplacian_of_gaussian_kernel_truncate
        self.log_pooling_method: None | str = laplacian_of_gaussian_pooling_method
        self.log_boundary_condition: None | str = laplacian_of_gaussian_boundary_condition

        # Check Laws kernel filter settings
        if self.has_laws_filter():
            # Check kernel.
            laws_kernel = self.check_laws_kernels(laws_kernel, "laws_kernel")

            # Check energy computation.
            if not isinstance(laws_compute_energy, bool):
                raise TypeError("The laws_compute_energy parameter is expected to be a boolean value.")

            if laws_compute_energy:

                # Check delta.
                if not isinstance(laws_delta, list):
                    laws_delta = [laws_delta]

                if not all(isinstance(delta, int) for delta in laws_delta):
                    raise TypeError(
                        "The laws_delta parameter is expected to be one or more integers with value 0 or "
                        "greater. Found: one or more values that are not integer.")

                if not all(delta >= 0 for delta in laws_delta):
                    raise ValueError(
                        "The laws_delta parameter is expected to be one or more integers with value 0 or "
                        "greater. Found: one or more values that are less than 0.")

            else:
                laws_delta = None

            # Check invariance.
            if not isinstance(laws_rotation_invariance, bool):
                raise TypeError("The laws_rotation_invariance parameter is expected to be a boolean value.")

            # Check pooling method.
            laws_pooling_method = self.check_pooling_method(laws_pooling_method, "laws_pooling_method")

            # Check boundary condition
            laws_boundary_condition = self.check_boundary_condition(laws_boundary_condition, "laws_boundary_condition")

        else:
            laws_kernel = None
            laws_compute_energy = None,
            laws_delta = None
            laws_rotation_invariance = None
            laws_pooling_method = None
            laws_boundary_condition = None

        self.laws_calculate_energy: None | bool = laws_compute_energy
        self.laws_kernel: None | list[str] = laws_kernel
        self.laws_delta: None | bool = laws_delta
        self.laws_rotation_invariance: None | bool = laws_rotation_invariance
        self.laws_pooling_method: None | str = laws_pooling_method
        self.laws_boundary_condition: None | str = laws_boundary_condition

        # Check Gabor filter settings.
        if self.has_gabor_filter():
            # Check sigma.
            gabor_sigma = self.check_sigma(gabor_sigma, "gabor_sigma")

            # Check gamma. Gamma behaves like sigma.
            gabor_gamma = self.check_sigma(gabor_gamma, "gabor_gamma")

            # Check lambda. Lambda behaves like sigma
            gabor_lambda = self.check_sigma(gabor_lambda, "gabor_lambda")

            # Check theta step.
            if gabor_theta_step is not None:
                if not isinstance(gabor_theta_step, (float, int)):
                    raise TypeError(
                        "The gabor_theta_step parameter is expected to be an angle, in degrees. Found a "
                        "value that was not a number.")

                if gabor_theta_step == 0.0:
                    gabor_theta_step = None

            if gabor_theta_step is not None:
                # Check that the step would divide the 360-degree circle into an integer number of steps.
                if not (360.0 / gabor_theta_step).is_integer():
                    raise ValueError(
                        f"The gabor_theta_step parameter should divide a circle into equal portions. "
                        f"The current settings would create {360.0 / gabor_theta_step} portions.")

            # Check theta.
            gabor_pool_theta = gabor_theta_step is not None

            if not isinstance(gabor_theta, list):
                gabor_theta = [gabor_theta]

            if gabor_theta_step is not None and len(gabor_theta) > 1:
                raise ValueError(
                    f"The gabor_theta parameter cannot have more than one value when used in conjunction"
                    f" with the gabor_theta_step parameter")

            if not all(isinstance(theta, (float, int)) for theta in gabor_theta):
                raise TypeError(
                    f"The gabor_theta parameter is expected to be one or more values indicating angles in"
                    f" degrees. Found: one or more values that were not numeric.")

            if gabor_theta_step is not None:
                gabor_theta = [gabor_theta[0] + ii for ii in np.arange(0.0, 360.0, gabor_theta_step)]

            # Check filter response.
            gabor_response = self.check_response(gabor_response, "gabor_response")

            # Check rotation invariance
            if not isinstance(gabor_rotation_invariance, bool):
                raise TypeError("The gabor_rotation_invariance parameter is expected to be a boolean value.")

            # Check pooling method
            gabor_pooling_method = self.check_pooling_method(gabor_pooling_method, "gabor_pooling_method")

            # Check boundary condition
            gabor_boundary_condition = self.check_boundary_condition(
                gabor_boundary_condition, "gabor_boundary_condition")

        else:
            gabor_sigma = None
            gabor_gamma = None
            gabor_lambda = None
            gabor_theta = None
            gabor_pool_theta = None
            gabor_response = None
            gabor_rotation_invariance = None
            gabor_pooling_method = None
            gabor_boundary_condition = None

        self.gabor_sigma: None | list[float] = gabor_sigma
        self.gabor_gamma: None | list[float] = gabor_gamma
        self.gabor_lambda: None | list[float] = gabor_lambda
        self.gabor_theta: None | list[float] | list[int] = gabor_theta
        self.gabor_pool_theta: None | bool = gabor_pool_theta
        self.gabor_response: None | str = gabor_response
        self.gabor_rotation_invariance: None | str = gabor_rotation_invariance
        self.gabor_pooling_method: None | str = gabor_pooling_method
        self.gabor_boundary_condition: None | str = gabor_boundary_condition

        # Check separable wavelet settings.
        if self.has_separable_wavelet_filter():
            # Check wavelet families.
            separable_wavelet_families = self.check_separable_wavelet_families(
                separable_wavelet_families, "separable_wavelet_families")

            # Check wavelet filter sets.
            separable_wavelet_set = self.check_separable_wavelet_sets(separable_wavelet_set, "separable_wavelet_set")

            # Check if wavelet is stationary
            if not isinstance(separable_wavelet_stationary, bool):
                raise TypeError(f"The separable_wavelet_stationary parameter is expected to be a boolean value.")

            # Check decomposition level
            separable_wavelet_decomposition_level = self.check_decomposition_level(
                separable_wavelet_decomposition_level, "separable_wavelet_decomposition_level")

            # Check rotation invariance
            if not isinstance(separable_wavelet_rotation_invariance, bool):
                raise TypeError("The separable_wavelet_rotation_invariance parameter is expected to be a boolean value.")

            # Check pooling method.
            separable_wavelet_pooling_method = self.check_pooling_method(
                separable_wavelet_pooling_method, "separable_wavelet_pooling_method")

            # Check boundary condition.
            separable_wavelet_boundary_condition = self.check_boundary_condition(
                separable_wavelet_boundary_condition, "separable_wavelet_boundary_condition")

        else:
            separable_wavelet_families = None
            separable_wavelet_set = None
            separable_wavelet_stationary = None
            separable_wavelet_decomposition_level = None
            separable_wavelet_rotation_invariance = None
            separable_wavelet_pooling_method = None
            separable_wavelet_boundary_condition = None

        self.separable_wavelet_families: None | list[str] = separable_wavelet_families
        self.separable_wavelet_filter_set: None | list[str] = separable_wavelet_set
        self.separable_wavelet_stationary: None | bool = separable_wavelet_stationary
        self.separable_wavelet_decomposition_level: None | list[int] = separable_wavelet_decomposition_level
        self.separable_wavelet_rotation_invariance: None | bool = separable_wavelet_rotation_invariance
        self.separable_wavelet_pooling_method: None | str = separable_wavelet_pooling_method
        self.separable_wavelet_boundary_condition: None | str = separable_wavelet_boundary_condition

        # Set parameters for non-separable wavelets.
        if self.has_nonseparable_wavelet_filter():
            # Check wavelet families.
            nonseparable_wavelet_families = self.check_nonseparable_wavelet_families(
                nonseparable_wavelet_families, "nonseparable_wavelet_families")

            # Check decomposition level.
            nonseparable_wavelet_decomposition_level = self.check_decomposition_level(
                nonseparable_wavelet_decomposition_level, "nonseparable_wavelet_decomposition_level")

            # Check filter response.
            nonseparable_wavelet_response = self.check_response(
                nonseparable_wavelet_response, "nonseparable_wavelet_response")

            # Check boundary condition.
            nonseparable_wavelet_boundary_condition = self.check_boundary_condition(
                nonseparable_wavelet_boundary_condition, "nonseparable_wavelet_boundary_condition")

        else:
            nonseparable_wavelet_families = None
            nonseparable_wavelet_decomposition_level = None
            nonseparable_wavelet_response = None
            nonseparable_wavelet_boundary_condition = None

        self.nonseparable_wavelet_families: None | list[str] = nonseparable_wavelet_families
        self.nonseparable_wavelet_decomposition_level: None | list[int] = nonseparable_wavelet_decomposition_level
        self.nonseparable_wavelet_response: None | str = nonseparable_wavelet_response
        self.nonseparable_wavelet_boundary_condition: None | str = nonseparable_wavelet_boundary_condition

        # Check Riesz filter orders.
        if self.has_riesz_filter():
            riesz_filter_order = self.check_riesz_filter_order(riesz_filter_order, "riesz_filter_order")

        else:
            riesz_filter_order = None

        if self.has_steered_riesz_filter() and self.ibsi_compliant:
            raise ValueError(
                "The steered riesz filters are not part of the IBSI reference standard. If you are sure that you want "
                "to use this method, use ibsi_compliant = False."
            )

        elif self.has_riesz_filter():
            riesz_filter_tensor_sigma = self.check_sigma(riesz_filter_tensor_sigma, "riesz_filter_tensor_sigma")

        else:
            riesz_filter_tensor_sigma = None

        self.riesz_order: None | list[list[int]] = riesz_filter_order
        self.riesz_filter_tensor_sigma: None | list[float] = riesz_filter_tensor_sigma

        if ibsi_compliant and self.has_square_transform_filter():
            raise ValueError(
                "The square transformation filter is not part of the IBSI reference standard. If you are sure that "
                "you want to use this method, use ibsi_compliant = False."
            )
        if ibsi_compliant and self.has_square_root_transform_filter():
            raise ValueError(
                "The square root transformation filter is not part of the IBSI reference standard. If you are sure "
                "that you want to use this method, use ibsi_compliant = False."
            )
        if ibsi_compliant and self.has_logarithm_transform_filter():
            raise ValueError(
                "The logarithmic transformation filter is not part of the IBSI reference standard. If you are sure "
                "that you want to use this method, use ibsi_compliant = False."
            )
        if ibsi_compliant and self.has_exponential_transform_filter():
            raise ValueError(
                "The exponential transformation filter is not part of the IBSI reference standard. If you are sure "
                "that you want to use this method, use ibsi_compliant = False."
            )

        self.lbp_separate_slices: None | list[bool] = None
        self.lbp_method: None | list[str] = None
        self.lbp_distance: None | list[float] = None

        if self.has_lbp_transform_filter():
            if ibsi_compliant:
                raise ValueError(
                    "The local binary pattern filter is not part of the IBSI reference standard. If you are sure "
                    "that you want to use this method, use ibsi_compliant = False."
                )

            lbp_separate_slices = []
            if any(filter_kernel in ["lbp", "lbp_2d"] for filter_kernel in self.spatial_filters):
                lbp_separate_slices += [True]
            elif any(filter_kernel in ["lbp_3d"] for filter_kernel in self.spatial_filters):
                lbp_separate_slices += [False]
            self.lbp_separate_slices = lbp_separate_slices

            # Check lbp_method.
            if not isinstance(lbp_method, str) and not isinstance(lbp_method, list):
                raise TypeError(f"The lbp_method parameter is expected to be a str or list of str.")
            if isinstance(lbp_method, str):
                lbp_method = [lbp_method]
            if not all(x in ["default", "variance", "rotation_invariant"] for x in lbp_method):
                raise ValueError(f"The lbp_method expects one or more of the following: default, variance, rotation_invariant")
            self.lbp_method = lbp_method

            # Check distance.
            if not isinstance(lbp_filter_distance, float) and not isinstance(lbp_filter_distance, list):
                raise TypeError(f"The lbp_filter_distance parameter is expected to be a float or list of float.")
            if isinstance(lbp_filter_distance, float):
                lbp_filter_distance = [lbp_filter_distance]
            if not all(x >= 1.0 for x in lbp_filter_distance):
                raise TypeError(f"lbp_filter_distance require a value of 1.0 or more.")
            self.lbp_distance = lbp_filter_distance

    @staticmethod
    def get_available_image_filters():
        return [
            "separable_wavelet", "nonseparable_wavelet", "riesz_nonseparable_wavelet",
            "riesz_steered_nonseparable_wavelet", "gaussian", "riesz_gaussian", "riesz_steered_gaussian",
            "laplacian_of_gaussian", "log", "riesz_laplacian_of_gaussian", "riesz_steered_laplacian_of_gaussian",
            "riesz_log", "riesz_steered_log", "laws", "gabor", "riesz_gabor", "riesz_steered_gabor", "mean",
            "pyradiomics_square", "pyradiomics_square_root", "pyradiomics_logarithm", "pyradiomics_exponential",
            "lbp", "lbp_2d", "lbp_3d"
        ]

    def check_boundary_condition(self, x, var_name):
        if x is None:
            if self.boundary_condition is not None:
                # Avoid updating by reference.
                x = copy.deepcopy(self.boundary_condition)

            else:
                raise ValueError(f"No value for the {var_name} parameter could be set, due to a lack of a default.")

        # Check value
        if x not in ["reflect", "constant", "nearest", "mirror", "wrap"]:
            raise ValueError(
                f"The provided value for the {var_name} is not valid. One of 'reflect', 'constant', "
                f"'nearest', 'mirror' or 'wrap' was expected. Found: {x}")

        return x

    @staticmethod
    def check_pooling_method(x, var_name, allow_none=False):

        valid_pooling_method = ["max", "min", "mean", "sum"]
        if allow_none:
            valid_pooling_method += ["none"]

        # Check if x is provided.
        if x is None:
            raise ValueError(
                f"No value for the {var_name} parameter could be found. One of the following values is expected: "
                f"{', '.join(valid_pooling_method)}"
            )

        # Check if x is a valid pooling method.
        if x not in valid_pooling_method:
            raise ValueError(
                f"The {var_name} parameter expects one of the following values: "
                f"{', '.join(valid_pooling_method)}. Found: {x}")

        return x

    @staticmethod
    def check_sigma(x, var_name):

        # Check if x is provided.
        if x is None:
            raise ValueError(
                f"No value for the {var_name} parameter could be found. One or more numerical values > 0.0 "
                f"are expected."
            )

        # Ensure that sigma is a list.
        if not isinstance(x, list):
            x = [x]

        # Check that the sigma values are floating points.
        if not all(isinstance(sigma, float) for sigma in x):
            raise TypeError(
                f"The {var_name} parameter is expected to consists of floating points with values "
                f"greater than 0.0. Found: one or more values that were not floating points.")

        if not all(sigma > 0.0 for sigma in x):
            raise ValueError(
                f"The {var_name} parameter is expected to consists of floating points with values "
                f"greater than 0.0. Found: one or more values with value 0.0 or less.")

        return x

    @staticmethod
    def check_truncation(x, var_name):

        # Check if x is provided.
        if x is None:
            raise ValueError(
                f"No value for the {var_name} parameter could be found. A numerical value > 0.0 is expected."
            )

        # Check that the truncation values are floating points.
        if not isinstance(x, float):
            raise TypeError(
                f"The {var_name} parameter is expected to be a floating point with value "
                f"greater than 0.0. Found: a value that was not a floating point.")

        if not x > 0.0:
            raise ValueError(
                f"The {var_name} parameter is expected to be a floating point with value "
                f"greater than 0.0. Found: a value of 0.0 or less.")

        return x

    @staticmethod
    def check_response(x, var_name):

        valid_response = ["modulus", "abs", "magnitude", "angle", "phase", "argument", "real", "imaginary"]

        # Check if x is provided.
        if x is None:
            raise ValueError(
                f"No value for the {var_name} parameter could be found. One of the following values is expected: "
                f"{', '.join(valid_response)}."
            )

        # Check that response is correct.
        if x not in valid_response:
            raise ValueError(
                f"The {var_name} parameter is not correct. Expected one of {', '.join(valid_response)}. "
                f"Found: {x}")

        return x

    @staticmethod
    def check_separable_wavelet_families(x, var_name):
        # Import pywavelets.
        import pywt

        # All available kernels from pywavelets.
        available_kernels = pywt.wavelist(kind="discrete")

        # Check if x is provided.
        if x is None:
            raise ValueError(
                f"No value for the {var_name} parameter could be found. One or more of the following values is "
                f"expected: {', '.join(available_kernels)}."
            )

        # Ensure that x is a list.
        if not isinstance(x, list):
            x = [x]

        # Check validity.
        valid_kernel = [kernel.lower() in available_kernels for kernel in x]
        if not all(valid_kernel):
            raise ValueError(
                f"The {var_name} parameter requires wavelet families that match those defined in the "
                f"pywavelets package. Could not match: "
                f"{', '.join([kernel for ii, kernel in x if not valid_kernel[ii]])}")

        # Return lowercase values.
        return [xx.lower() for xx in x]

    @staticmethod
    def check_nonseparable_wavelet_families(x, var_name):

        available_kernels = ["simoncelli", "shannon"]

        # Check if x is provided.
        if x is None:
            raise ValueError(
                f"No value for the {var_name} parameter could be found. One or more of the following values is "
                f"expected: {', '.join(available_kernels)}."
            )

        # Ensure that x is a list.
        if not isinstance(x, list):
            x = [x]

        # Check validity.
        valid_kernel = [kernel.lower() in available_kernels for kernel in x]
        if not all(valid_kernel):
            raise ValueError(
                f"The {var_name} parameter expects one or more of the following values: "
                f"{', '.join(available_kernels)}. Could not match: "
                f"{', '.join([kernel for ii, kernel in x if not valid_kernel[ii]])}")

        # Return lowercase values.
        return [xx.lower() for xx in x]

    @staticmethod
    def check_decomposition_level(x, var_name):

        # Check if x is provided.
        if x is None:
            raise ValueError(
                f"No value for the {var_name} parameter could be found. One or more integer values equal to or greater "
                f"than 1 are expected."
            )

        # Ensure that x is a list.
        if not isinstance(x, list):
            x = [x]

        # Check that all values in x are integers.
        if not all(isinstance(xx, int) for xx in x):
            raise TypeError(
                f"The {var_name} parameter should be one or more integer "
                f"values of at least 1. Found: one or more values that was not an integer.")

        # Check that all values in x are greater or equal to 1.
        if not all(xx >= 1 for xx in x):
            raise ValueError(
                f"The {var_name} parameter should be one or more integer "
                f"values of at least 1. Found: one or more values that was not an integer.")

        return x

    def check_separable_wavelet_sets(self, x: None | str | list[str], var_name: str):
        from itertools import product

        # Set default value.
        if x is None:
            if self.by_slice:
                x = "hh"
            else:
                x = "hhh"

        # Check if x is a list.
        if not isinstance(x, list):
            x = [x]

        # Generate all potential combinations.
        if self.by_slice:
            possible_combinations = ["".join(combination) for combination in product(["l", "h"], repeat=2)]

        else:
            possible_combinations = ["".join(combination) for combination in product(["l", "h"], repeat=3)]

        # Check for all.
        if any(kernel == "all" for kernel in x):
            x = possible_combinations

        # Check which kernels are valid.
        valid_kernel = [kernel.lower() in possible_combinations for kernel in x]

        if not all(valid_kernel):
            raise ValueError(
                f"The {var_name} parameter requires combinations of low (l) and high-pass (h) kernels. "
                f"Two kernels should be specified for 2D, and three for 3D. Found the following invalid "
                f"combinations: "
                f"{', '.join([kernel for ii, kernel in enumerate(x) if not valid_kernel[ii]])}")

        # Return lowercase values.
        return [xx.lower() for xx in x]

    def check_laws_kernels(self, x: None | str | list[str], var_name: str):
        from itertools import product

        # Set implemented kernels.
        kernels = ['l5', 'e5', 's5', 'w5', 'r5', 'l3', 'e3', 's3']

        # Generate all valid combinations.
        if self.by_slice:
            possible_combinations = ["".join(combination) for combination in product(kernels, repeat=2)]

        else:
            possible_combinations = ["".join(combination) for combination in product(kernels, repeat=3)]

        # Check if x is provided.
        if x is None:
            raise ValueError(
                f"No value for the {var_name} parameter could be found. One or more of the following values is "
                f"expected: {', '.join(possible_combinations)}."
            )

        # Ensure that x is a list.
        if not isinstance(x, list):
            x = [x]

        # Check which kernels are valid.
        valid_kernel = [kernel.lower() in possible_combinations for kernel in x]
        if not all(valid_kernel):
            raise ValueError(
                f"The {var_name} parameter requires combinations of Laws kernels. The follow kernels are "
                f"implemented: {', '.join(kernels)}. Two kernels should be specified for 2D, "
                f"and three for 3D. Found the following illegal combinations: "
                f"{', '.join([kernel for ii, kernel in enumerate(x) if not valid_kernel[ii]])}"
            )

        # Return lowercase values.
        return [xx.lower() for xx in x]

    def check_riesz_filter_order(self, x, var_name):
        from itertools import product

        # Skip if None
        if x is None:
            return x

        # Set number of elements that the filter order should have
        if self.by_slice:
            n_elements = 2

        else:
            n_elements = 3

        # Create filterbank.
        if isinstance(x, int):
            # Check that x is not negative.
            if x < 0:
                raise ValueError(f"The {var_name} parameter cannot be negative.")

            # Set filter order.
            single_filter_order = list(range(x+1))

            # Generate all valid combinations.
            x = [list(combination) for combination in product(single_filter_order, repeat=n_elements) if
                 sum(combination) == x]

        if not isinstance(x, list):
            raise TypeError(f"The {var_name} parameter is expected to be a list")

        # Create a nested list,
        if not all(isinstance(xx, list) for xx in x):
            x = [x]

        # Check that all elements of x have the right length, and do not negative orders.
        if not all(len(xx) == n_elements for xx in x):
            raise ValueError(
                f"The {var_name} parameter is expected to contain filter orders, each consisting of "
                f"{n_elements} non-negative integer values. One or more filter orders did not have the "
                f"expected number of elements.")

        if not all(all(isinstance(xxx, int) for xxx in xx) for xx in x):
            raise ValueError(
                f"The {var_name} parameter is expected to contain filter orders, each consisting of "
                f"{n_elements} non-negative integer values. One or more filter orders did not fully "
                f"consist of integer values.")

        if not all(all(xxx >= 0 for xxx in xx) for xx in x):
            raise ValueError(
                f"The {var_name} parameter is expected to contain filter orders, each consisting of "
                f"{n_elements} non-negative integer values. One or more filter orders contained negative values.")

        return x

    def has_mean_filter(self, x=None):
        if x is None:
            x = self.spatial_filters
        elif not isinstance(x, list):
            x = [x]

        return x is not None and any(filter_kernel in ["mean"] for filter_kernel in x)

    def has_gaussian_filter(self, x=None):
        if x is None:
            x = self.spatial_filters
        elif not isinstance(x, list):
            x = [x]

        return x is not None and any(
            filter_kernel in ["gaussian", "riesz_gaussian", "riesz_steered_gaussian"] for filter_kernel in x)

    def has_laplacian_of_gaussian_filter(self, x=None):
        if x is None:
            x = self.spatial_filters
        elif not isinstance(x, list):
            x = [x]

        return x is not None and any(
            filter_kernel in [
                "laplacian_of_gaussian", "log", "riesz_laplacian_of_gaussian", "riesz_log",
                "riesz_steered_laplacian_of_gaussian", "riesz_steered_log"
            ] for filter_kernel in x)

    def has_laws_filter(self, x=None):
        if x is None:
            x = self.spatial_filters
        elif not isinstance(x, list):
            x = [x]

        return x is not None and any(filter_kernel in ["laws"] for filter_kernel in x)

    def has_gabor_filter(self, x=None):
        if x is None:
            x = self.spatial_filters
        elif not isinstance(x, list):
            x = [x]

        return x is not None and any(
            filter_kernel in ["gabor", "riesz_gabor", "riesz_steered_gabor"] for filter_kernel in x)

    def has_separable_wavelet_filter(self, x=None):
        if x is None:
            x = self.spatial_filters
        elif not isinstance(x, list):
            x = [x]

        return x is not None and any(filter_kernel in ["separable_wavelet"] for filter_kernel in x)

    def has_nonseparable_wavelet_filter(self, x=None):
        if x is None:
            x = self.spatial_filters
        elif not isinstance(x, list):
            x = [x]

        return x is not None and any(
            filter_kernel in [
                "nonseparable_wavelet", "riesz_nonseparable_wavelet", "riesz_steered_nonseparable_wavelet"
            ] for filter_kernel in x)

    def has_riesz_filter(self, x=None):
        if x is None:
            x = self.spatial_filters
        elif not isinstance(x, list):
            x = [x]

        return x is not None and any(filter_kernel.startswith("riesz") for filter_kernel in x)

    def has_steered_riesz_filter(self, x=None):
        if x is None:
            x = self.spatial_filters
        elif not isinstance(x, list):
            x = [x]

        return x is not None and any(filter_kernel.startswith("riesz_steered") for filter_kernel in x)

    def has_square_transform_filter(self, x=None):
        if x is None:
            x = self.spatial_filters
        elif not isinstance(x, list):
            x = [x]

        return x is not None and any(filter_kernel == "pyradiomics_square" for filter_kernel in x)

    def has_square_root_transform_filter(self, x=None):
        if x is None:
            x = self.spatial_filters
        elif not isinstance(x, list):
            x = [x]

        return x is not None and any(filter_kernel == "pyradiomics_square_root" for filter_kernel in x)

    def has_logarithm_transform_filter(self, x=None):
        if x is None:
            x = self.spatial_filters
        elif not isinstance(x, list):
            x = [x]

        return x is not None and any(filter_kernel == "pyradiomics_logarithm" for filter_kernel in x)

    def has_exponential_transform_filter(self, x=None):
        if x is None:
            x = self.spatial_filters
        elif not isinstance(x, list):
            x = [x]

        return x is not None and any(filter_kernel == "pyradiomics_exponential" for filter_kernel in x)

    def has_lbp_transform_filter(self, x=None):
        if x is None:
            x = self.spatial_filters
        elif not isinstance(x, list):
            x = [x]

        return x is not None and any(filter_kernel in ["lbp", "lbp_2d", "lbp_3d"] for filter_kernel in x)

def get_image_transformation_settings() -> list[dict[str, Any]]:
    return [
        setting_def(
            "response_map_feature_families", "str", to_list=True, xml_key="feature_families",
            class_key="families", test=["statistical", "glcm"]
        ),
        setting_def(
            "response_map_discretisation_method", "str", to_list=True, xml_key="discretisation_method",
            class_key="discretisation_method", test=["fixed_bin_size", "fixed_bin_number"]
        ),
        setting_def(
            "response_map_discretisation_n_bins", "int", to_list=True, xml_key="discretisation_n_bins",
            class_key="discretisation_n_bins", test=[10, 33]
        ),
        setting_def(
            "response_map_discretisation_bin_width", "float", to_list=True, xml_key="discretisation_bin_width",
            class_key="discretisation_bin_width", test=[10.0, 34.0]
        ),
        setting_def(
            "filter_kernels", "str", to_list=True, xml_key=["filter_kernels", "spatial_filters"],
            class_key="spatial_filters", test=[
                "separable_wavelet", "nonseparable_wavelet", "riesz_nonseparable_wavelet", "gaussian", "riesz_gaussian",
                "laplacian_of_gaussian", "log", "riesz_laplacian_of_gaussian", "riesz_log", "laws", "gabor",
                "riesz_gabor", "mean",
                "lbp_3d"
            ]
        ),
        setting_def("boundary_condition", "str", test="nearest"),
        setting_def("separable_wavelet_families", "str", to_list=True, test=["coif4", "coif5"]),
        setting_def(
            "separable_wavelet_set", "str", to_list=True, class_key="separable_wavelet_filter_set",
            test=["hhh", "lll"]
        ),
        setting_def("separable_wavelet_stationary", "bool", test=False),
        setting_def("separable_wavelet_decomposition_level", "int", to_list=True, test=[1, 2]),
        setting_def("separable_wavelet_rotation_invariance", "bool", test=False),
        setting_def("separable_wavelet_pooling_method", "str", test="mean"),
        setting_def("separable_wavelet_boundary_condition", "str", test="constant"),
        setting_def("nonseparable_wavelet_families", "str", to_list=True, test=["simoncelli", "shannon"]),
        setting_def("nonseparable_wavelet_decomposition_level", "int", to_list=True, test=[1, 2]),
        setting_def("nonseparable_wavelet_response", "str", test="magnitude"),
        setting_def("nonseparable_wavelet_boundary_condition", "str", test="constant"),
        setting_def("gaussian_sigma", "float", to_list=True, test=[1.0, 3.0]),
        setting_def("gaussian_kernel_truncate", "float", class_key="gaussian_sigma_truncate", test=10.0),
        setting_def(
            "gaussian_kernel_boundary_condition", "str", class_key="gaussian_boundary_condition", test="constant"
        ),
        setting_def(
            "laplacian_of_gaussian_sigma", "float", to_list=True,
            xml_key=["laplacian_of_gaussian_sigma", "log_sigma"], class_key="log_sigma", test=[1.0, 3.0]
        ),
        setting_def(
            "laplacian_of_gaussian_kernel_truncate", "float",
            xml_key=["laplacian_of_gaussian_kernel_truncate", "log_sigma_truncate"], class_key="log_sigma_truncate",
            test=10.0
        ),
        setting_def("laplacian_of_gaussian_pooling_method", "str", class_key="log_pooling_method", test="mean"),
        setting_def(
            "laplacian_of_gaussian_boundary_condition", "str", class_key="log_boundary_condition", test="constant"
        ),
        setting_def("laws_kernel", "str", to_list=True, test=["l5e5s5", "w5r5l3"]),
        setting_def(
            "laws_compute_energy", "bool", xml_key="laws_calculate_energy",
            class_key="laws_calculate_energy", test=True
        ),
        setting_def("laws_delta", "int", to_list=True, test=[3, 5]),
        setting_def(
            "laws_rotation_invariance", "bool", xml_key=["laws_rotation_invariance", "laws_rot_invar"], test=False
        ),
        setting_def("laws_pooling_method", "str", test="mean"),
        setting_def("laws_boundary_condition", "str", test="constant"),
        setting_def("gabor_sigma", "float", to_list=True, test=[1.0, 3.0]),
        setting_def("gabor_lambda", "float", to_list=True, test=[0.5, 2.0]),
        setting_def("gabor_gamma", "float", to_list=True, test=[0.5, 0.75]),
        setting_def("gabor_theta", "float", to_list=True, test=[5.0, 15.0]),
        setting_def("gabor_theta_step", "float", test=None),
        setting_def("gabor_response", "str", test="magnitude"),
        setting_def(
            "gabor_rotation_invariance", "bool", xml_key=["gabor_rotation_invariance", "gabor_rot_invar"], test=False
        ),
        setting_def("gabor_pooling_method", "str", test="mean"),
        setting_def("gabor_boundary_condition", "str", test="constant"),
        setting_def(
            "mean_filter_kernel_size", "int", to_list=True, xml_key=["mean_filter_kernel_size", "mean_filter_size"],
            class_key="mean_filter_size", test=[3, 7]
        ),
        setting_def("mean_filter_boundary_condition", "str", test="constant"),
        setting_def(
            "riesz_filter_order", "int", to_list=True, xml_key=["riesz_filter_order", "riesz_order"],
            class_key="riesz_order", test=[2, 1, 0]
        ),
        setting_def("riesz_filter_tensor_sigma", "float", to_list=True, test=[3.0, 5.0]),
        setting_def("lbp_method", "str", to_list=True, test=["variance"]),
        setting_def("lbp_filter_distance", "float", to_list=True, class_key="lbp_distance", test=[2.0, 3.0])
    ]



from functools import lru_cache
from typing import Generator

import numpy as np
import pandas as pd

from mirp._images.generic_image import GenericImage
from mirp._masks.base_mask import BaseMask
from mirp._features.histogram import get_discretisation_parameters, HistogramDerivedFeature
from mirp.settings.feature_parameters import FeatureExtractionSettingsClass


class DataIH(object):

    def __init__(self):
        # Raw intensity data
        self.image: np.ndarray | None = None

        # Histogram data
        self.histogram: pd.DataFrame | None = None

        # Number of voxels
        self.n_voxels: int | None = None

        # Number of grey levels
        self.n_g: int | None = None

        # Mean intensity
        self.mu: float | None = None

        # Intensity standard deviation
        self.sigma: float | None = None

    def compute(self, image: GenericImage, mask: BaseMask):
        # Skip processing if input image and/or roi are missing
        if image is None or mask is None:
            return

        # Check if data actually exists
        if image.is_empty() or mask.roi_intensity.is_empty_mask():
            return

        # Convert to dataframe and remove entries outside the mask.
        image = mask.as_pandas_dataframe(image=image, intensity_mask=True)
        image = image[image.roi_int_mask == True]
        if len(image) == 0:
            return

        # Set image data
        self.image = image.g.values

        # Number of voxels
        self.n_voxels = len(image)

        # Number of grey levels
        self.n_g = mask.intensity_range[1] - mask.intensity_range[0] + 1

        # Define histogram
        histogram = image.groupby(by="g").size().reset_index(name="n")

        # Append empty grey levels to histogram
        levels = np.arange(start=0, stop=self.n_g) + 1
        missing_level = levels[np.logical_not(np.isin(levels, histogram.g))]
        n_missing = len(missing_level)
        if n_missing > 0:
            histogram = pd.concat([
                histogram,
                pd.DataFrame({
                    "g": missing_level,
                    "n": np.zeros(n_missing)
                })
            ], ignore_index=True)

        # Update histogram by sorting grey levels and adding bin probabilities
        histogram = histogram.sort_values(by="g")
        histogram["p"] = histogram.n / self.n_voxels

        # Add gradient data.
        if len(histogram) > 1:
            histogram["grad"] = np.gradient(histogram.n)
        else:
            histogram["grad"] = 0.0

        self.histogram = histogram

        # Mean intensity
        self.mu = np.sum(histogram.g * histogram.p)

        # Standard deviation
        self.sigma = np.sqrt(np.sum((histogram.g - self.mu) ** 2.0 * histogram.p))

    def is_empty(self):
        return self.image is None or self.histogram is None


class FeatureIH(HistogramDerivedFeature):

    def __init__(
            self,
            **kwargs
    ):
        super().__init__(**kwargs)

        # Perform close crop for intensity histograms.
        self.cropping_distance = 0.0

    @staticmethod
    @lru_cache(maxsize=1)
    def _get_data(
            image: GenericImage,
            mask: BaseMask
    ) -> DataIH:
        data = DataIH()
        data.compute(image=image, mask=mask)

        return data

    def clear_local_cache(self, other):
        if not isinstance(other, FeatureIH):
            self._get_data.cache_clear()

    def clear_cache(self):
        super().clear_cache()
        self._get_data.cache_clear()

    def compute(self, image: GenericImage, mask: BaseMask):
        # Discretise images.
        image, mask = self.discretise_image(
            image=image,
            mask=mask,
            discretisation_method=self.discretisation_method,
            bin_width=self.bin_width,
            bin_number=self.bin_number,
            cropping_distance=self.cropping_distance
        )

        # Get data.
        data = self._get_data(image=image, mask=mask)

        # Compute feature value.
        if data.is_empty():
            self.value = np.nan
        else:
            self.value = self._compute(data=data)

    @staticmethod
    def _compute(data: DataIH):
        raise NotImplementedError("Implement _compute for feature-specific computation.")

    def create_table_name(self):
        table_elements = (
                self._get_base_table_name_element()
                + self._get_discretisation_table_name_element()
        )
        self.table_name = "_".join(table_elements)


class FeatureIHMean(FeatureIH):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "IH - mean"
        self.abbr_name = "ih_mean"
        self.ibsi_id = "X6K6"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: DataIH) -> float:
        return data.mu


class FeatureIHVariance(FeatureIH):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "IH - variance"
        self.abbr_name = "ih_var"
        self.ibsi_id = "CH89"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: DataIH) -> float:
        return data.sigma ** 2.0


class FeatureIHSkewness(FeatureIH):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "IH - skewness"
        self.abbr_name = "ih_skew"
        self.ibsi_id = "88K1"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: DataIH) -> float:
        if data.sigma == 0.0:
            return 0.0
        return np.sum((data.histogram.g - data.mu) ** 3.0 * data.histogram.p) / (data.sigma ** 3.0)


class FeatureIHKurtosis(FeatureIH):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "IH - kurtosis"
        self.abbr_name = "ih_kurt"
        self.ibsi_id = "C3I7"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: DataIH) -> float:
        if data.sigma == 0.0:
            return 0.0
        return np.sum((data.histogram.g - data.mu) ** 4.0 * data.histogram.p) / (data.sigma ** 4.0) - 3.0


class FeatureIHMedian(FeatureIH):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "IH - median"
        self.abbr_name = "ih_median"
        self.ibsi_id = "WIFQ"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: DataIH) -> float:
        return np.median(data.image)


class FeatureIHMinimum(FeatureIH):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "IH - minimum"
        self.abbr_name = "ih_min"
        self.ibsi_id = "1PR8"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: DataIH) -> float:
        return np.min(data.image)


class FeatureIHPercentile(FeatureIH):
    def __init__(self, percentile: float, **kwargs):
        # This class allows for passing percentile values.
        super().__init__(**kwargs)

        # Set percentile
        self.percentile = percentile

        # Set IBSI identifier. Only the 10th and 90th percentile are explicitly identified.
        ibsi_id = ""
        if percentile == 10.0:
            ibsi_id = "GPMT"
        elif percentile == 90.0:
            ibsi_id = "OZ0C"
        self.ibsi_id = ibsi_id

        if percentile.is_integer():
            self.name = f"IH - {int(percentile)}th percentile"
            self.abbr_name = f"ih_p{int(percentile)}"
        else:
            self.name = f"IH - {percentile}th percentile"
            self.abbr_name = f"ih_p{percentile}"
        self.ibsi_compliant = True

    def compute(self, image: GenericImage, mask: BaseMask):
        # Discretise images.
        image, mask = self.discretise_image(
            image=image,
            mask=mask,
            discretisation_method=self.discretisation_method,
            bin_width=self.bin_width,
            bin_number=self.bin_number,
            cropping_distance=self.cropping_distance
        )

        # Get data.
        data = self._get_data(image=image, mask=mask)

        # Compute feature value.
        if data.is_empty():
            self.value = np.nan
        else:
            self.value = np.percentile(data.image, q=self.percentile)


class FeatureIHMaximum(FeatureIH):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "IH - maximum"
        self.abbr_name = "ih_max"
        self.ibsi_id = "3NCY"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: DataIH) -> float:
        return np.max(data.image)


class FeatureIHMode(FeatureIH):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "IH - mode"
        self.abbr_name = "ih_mode"
        self.ibsi_id = "AMMC"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: DataIH) -> float:
        mode_g = data.histogram.loc[data.histogram.n == np.max(data.histogram.n)].g.values
        return mode_g[np.argmin(np.abs(mode_g - data.mu))]


class FeatureIHInterQuartileRange(FeatureIH):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "IH - interquartile range"
        self.abbr_name = "ih_iqr"
        self.ibsi_id = "WR0O"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: DataIH) -> float:
        return np.percentile(data.image, q=75) - np.percentile(data.image, q=25)


class FeatureIHRange(FeatureIH):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "IH - range"
        self.abbr_name = "ih_range"
        self.ibsi_id = "5Z3W"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: DataIH) -> float:
        return float(np.max(data.image) - np.min(data.image))


class FeatureIHMeanAbsoluteDeviation(FeatureIH):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "IH - mean absolute deviation"
        self.abbr_name = "ih_mad"
        self.ibsi_id = "D2ZX"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: DataIH) -> float:
        return np.mean(np.abs(data.image - data.mu))


class FeatureIHRobustMeanAbsoluteDeviation(FeatureIH):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "IH - robust mean absolute deviation"
        self.abbr_name = "ih_rmad"
        self.ibsi_id = "WRZB"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: DataIH) -> float:
        selected_values = data.image[
            (data.image >= np.percentile(data.image, q=10))
            & (data.image <= np.percentile(data.image, q=90))
        ]
        return np.mean(np.abs(selected_values - np.mean(selected_values)))


class FeatureIHMedianAbsoluteDeviation(FeatureIH):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "IH - median absolute deviation"
        self.abbr_name = "ih_medad"
        self.ibsi_id = "4RNL"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: DataIH) -> float:
        return np.mean(np.abs(data.image - np.median(data.image)))


class FeatureIHCoefficientOfVariation(FeatureIH):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "IH - coefficient of variation"
        self.abbr_name = "ih_cov"
        self.ibsi_id = "CWYJ"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: DataIH) -> float:
        if data.sigma == 0.0:
            return 0.0
        return data.sigma / data.mu


class FeatureIHQuartileCoefficientOfDispersion(FeatureIH):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "IH - quartile coefficient of dispersion"
        self.abbr_name = "ih_qcod"
        self.ibsi_id = "SLWD"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: DataIH) -> float:
        denom = np.percentile(data.image, q=75) + np.percentile(data.image, q=25)
        if denom == 0.0:
            return 1.0

        return (np.percentile(data.image, q=75) - np.percentile(data.image, q=25)) / denom


class FeatureIHEntropy(FeatureIH):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "IH - entropy"
        self.abbr_name = "ih_entropy"
        self.ibsi_id = "TLU2"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: DataIH) -> float:
        return -np.sum(data.histogram.p[data.histogram.p > 0.0] * np.log2(data.histogram.p[data.histogram.p > 0.0]))


class FeatureIHUniformity(FeatureIH):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "IH - uniformity"
        self.abbr_name = "ih_uniformity"
        self.ibsi_id = "BJ5W"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: DataIH) -> float:
        return np.sum(data.histogram.p ** 2.0)


class FeatureIHMaximumHistogramGradient(FeatureIH):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "IH - maximum histogram gradient"
        self.abbr_name = "ih_max_grad"
        self.ibsi_id = "12CE"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: DataIH) -> float:
        return np.max(data.histogram.grad)


class FeatureIHMaximumHistogramGradientGreyLevel(FeatureIH):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "IH - maximum histogram gradient grey level"
        self.abbr_name = "ih_max_grad_g"
        self.ibsi_id = "8E6O"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: DataIH) -> float:
        return data.histogram.g[data.histogram.grad.idxmax()]


class FeatureIHMinimumHistogramGradient(FeatureIH):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "IH - minimum histogram gradient"
        self.abbr_name = "ih_min_grad"
        self.ibsi_id = "VQB3"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: DataIH) -> float:
        return np.min(data.histogram.grad)


class FeatureIHMinimumHistogramGradientGreyLevel(FeatureIH):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "IH - minimum histogram gradient grey level"
        self.abbr_name = "ih_min_grad_g"
        self.ibsi_id = "RHQZ"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: DataIH) -> float:
        return data.histogram.g[data.histogram.grad.idxmin()]


def get_ih_class_dict() -> dict[str, FeatureIH]:
    class_dict = {
        "ih_mean": FeatureIHMean,
        "ih_var": FeatureIHVariance,
        "ih_skew": FeatureIHSkewness,
        "ih_kurt": FeatureIHKurtosis,
        "ih_median": FeatureIHMedian,
        "ih_min": FeatureIHMinimum,
        "ih_p": FeatureIHPercentile,
        "ih_max": FeatureIHMaximum,
        "ih_mode": FeatureIHMode,
        "ih_iqr": FeatureIHInterQuartileRange,
        "ih_range": FeatureIHRange,
        "ih_mad": FeatureIHMeanAbsoluteDeviation,
        "ih_rmad": FeatureIHRobustMeanAbsoluteDeviation,
        "ih_medad": FeatureIHMedianAbsoluteDeviation,
        "ih_cov": FeatureIHCoefficientOfVariation,
        "ih_qcod": FeatureIHQuartileCoefficientOfDispersion,
        "ih_entropy": FeatureIHEntropy,
        "ih_uniformity": FeatureIHUniformity,
        "ih_max_grad": FeatureIHMaximumHistogramGradient,
        "ih_max_grad_g": FeatureIHMaximumHistogramGradientGreyLevel,
        "ih_min_grad": FeatureIHMinimumHistogramGradient,
        "ih_min_grad_g": FeatureIHMinimumHistogramGradientGreyLevel
    }

    return class_dict


def generate_ih_features(
        settings: FeatureExtractionSettingsClass,
        features: None | list[str] = None
) -> Generator[FeatureIH, None, None]:
    class_dict = get_ih_class_dict()
    ih_features = list(class_dict.keys())

    # Populate features if available.
    if features is None and settings.has_ih_family():
        features = ih_features

    # Terminate early if no features are set, and none are required.
    if features is None:
        return

    # Select only DZM-features, and return if none are present.
    features = [feature for feature in features if feature in ih_features]
    if len(features) == 0:
        return

    # Set default percentiles.
    percentiles = [10.0, 90.0]

    # Features are parametrised by the choice of discretisation parameters and spatial methods.
    for discretisation_parameters in get_discretisation_parameters(
        settings=settings
    ):
        for feature in features:
            if feature == "ih_p":
                for percentile in percentiles:
                    yield class_dict[feature](
                        percentile=percentile,
                        **discretisation_parameters
                    )
            else:
                yield class_dict[feature](
                    **discretisation_parameters
                )



from typing import Any
from dataclasses import dataclass
from mirp.settings.utilities import setting_def

import numpy as np


@dataclass
class ResegmentationSettingsClass:
    """
    Parameters related to mask resegmentation. Resegmentation is used to remove parts of the mask that correspond to
    undesired intensities that should be excluded, e.g. those corresponding to air. Resegmentation based on an
    intensity range is also required for using *Fixed Bin Size* discretisation to set the lower bound of the first bin.

    .. note::
        Even though intensity range resegmentation is usually required to perform *Fixed Bin Size* discretisation,
        default values are available for computed tomography (CT) and positron emission tomography (PET) imaging,
        and are set to -1000.0 Hounsfield Units and 0.0 SUV, respectively.

    Parameters
    ----------
    resegmentation_intensity_range: list of float, optional
        Intensity threshold for threshold-based re-segmentation ("threshold" and "range"). If set, requires two
        values for lower and upper range respectively. The upper range value can also be np.nan for half-open ranges.

    resegmentation_sigma: float, optional
        Number of standard deviations for outlier-based intensity re-segmentation ("sigma" and "outlier").

    **kwargs: dict, optional
        Unused keyword arguments.
    """
    def __init__(
            self,
            resegmentation_intensity_range: None | list[float] = None,
            resegmentation_sigma: None | float = None,
            **kwargs
    ):
        resegmentation_method = []
        if resegmentation_sigma is None and resegmentation_intensity_range is None:
            resegmentation_method += ["none"]
        if resegmentation_intensity_range is not None:
            resegmentation_method += ["range"]
        if resegmentation_sigma is not None:
            resegmentation_method += ["sigma"]

        # Set resegmentation method.
        self.resegmentation_method: list[str] = resegmentation_method

        # Set default value.
        if resegmentation_intensity_range is None:
            # Cannot define a proper range.
            resegmentation_intensity_range = [np.nan, np.nan]

        if not isinstance(resegmentation_intensity_range, list):
            raise TypeError(
                f"The resegmentation_intensity_range parameter should be a list with exactly two "
                f"values. Found: an object that is not a list.")

        if len(resegmentation_intensity_range) != 2:
            raise ValueError(
                f"The resegmentation_intensity_range parameter should be a list with exactly two "
                f"values. Found: list with {len(resegmentation_intensity_range)} values.")

        if not all(isinstance(ii, float) for ii in resegmentation_intensity_range):
            raise TypeError(
                f"The resegmentation_intensity_range parameter should be a list with exactly two "
                f"values. Found: one or more values that are not floating point values.")

        self.intensity_range: None | tuple[Any, Any] = tuple(resegmentation_intensity_range) if \
            resegmentation_intensity_range is not None else None

        # Set default value.
        if resegmentation_sigma is None:
            resegmentation_sigma = 3.0

        # Check that sigma is not negative.
        if resegmentation_sigma < 0.0:
            raise ValueError(f"The resegmentation_sigma parameter can not be negative. Found: {resegmentation_sigma}")

        self.sigma: float = resegmentation_sigma


def get_mask_resegmentation_settings() -> list[dict[str, Any]]:
    return [
        setting_def(
            "resegmentation_intensity_range", "float", to_list=True, xml_key=["intensity_range", "g_thresh"],
            class_key="intensity_range", test=[-10.0, 30.0]
        ),
        setting_def("resegmentation_sigma", "float", xml_key="sigma", class_key="sigma", test=1.0)
    ]



from functools import lru_cache
from typing import Generator

import numpy as np

from mirp._features.ngtdm_matrix import MatrixNGTDM
from mirp._images.generic_image import GenericImage
from mirp._masks.base_mask import BaseMask
from mirp._features.histogram import get_discretisation_parameters
from mirp._features.texture_features import FeatureTexture
from mirp.settings.feature_parameters import FeatureExtractionSettingsClass


class FeatureNGTDM(FeatureTexture):

    def __init__(
            self,
            **kwargs
    ):
        super().__init__(**kwargs)

        # Perform close crop for NGTDM.
        self.cropping_distance = 0.0

    def get_matrix(
            self,
            image: GenericImage,
            mask: BaseMask,
    ) -> list[MatrixNGTDM]:
        # First discretise the image.
        image, mask = self.discretise_image(
            image=image,
            mask=mask,
            discretisation_method=self.discretisation_method,
            bin_width=self.bin_width,
            bin_number=self.bin_number,
            cropping_distance=self.cropping_distance
        )

        # Then get matrix or matrices
        matrix_list = self._get_matrix(
            image=image,
            mask=mask,
            spatial_method=self.spatial_method
        )

        return matrix_list

    @staticmethod
    @lru_cache(maxsize=1)
    def _get_matrix(
            image: GenericImage,
            mask: BaseMask,
            spatial_method: str
    ) -> list[MatrixNGTDM]:
        # Represent image and mask as a dataframe.
        data = mask.as_pandas_dataframe(
            image=image,
            intensity_mask=True
        )

        # Instantiate a helper copy of the current class to be able to use class methods without tying the cache to the
        # instance of the original object from which this method is called.
        matrix_instance = MatrixNGTDM(
            spatial_method=spatial_method
        )

        # Compute the required matrices.
        matrix_list = list(matrix_instance.generate(prototype=MatrixNGTDM, n_slices=image.image_dimension[0]))
        for matrix in matrix_list:
            matrix.compute(data=data, image_dimension=image.image_dimension)

        # Merge according to the spatial method.
        matrix_list = matrix_instance.merge(matrix_list, prototype=MatrixNGTDM)

        # Compute additional values from the individual matrices.
        for matrix in matrix_list:
            matrix.set_values_from_matrix(intensity_range=mask.intensity_range)

        return matrix_list

    def clear_local_cache(self, other):
        if not isinstance(other, FeatureNGTDM):
            self._get_matrix.cache_clear()

    def clear_cache(self):
        super().clear_cache()
        self._get_matrix.cache_clear()

    def compute(self, image: GenericImage, mask: BaseMask):
        # Skip processing if input image and/or roi are missing
        if image is None or mask is None:
            return

        # Check if data actually exists
        if image.is_empty() or mask.roi_intensity.is_empty_mask():
            return

        # Compute or retrieve matrices from cache.
        matrices = self.get_matrix(image=image, mask=mask)

        # Compute feature value from matrices, and average over matrices.
        values = [self._compute(matrix=matrix) for matrix in matrices]
        if np.all(np.isnan(values)):
            self.value = np.nan
        else:
            self.value = np.nanmean(values)

    @staticmethod
    def _compute(matrix: MatrixNGTDM):
        raise NotImplementedError("Implement _compute for feature-specific computation.")

    def create_table_name(self):
        table_elements = (
                self._get_base_table_name_element()
                + self._get_spatial_table_name_element()
                + self._get_discretisation_table_name_element()
        )
        self.table_name = "_".join(table_elements)


class FeatureNGTDMCoarseness(FeatureNGTDM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "NGTDM - coarseness"
        self.abbr_name = "ngt_coarseness"
        self.ibsi_id = "QCDE"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixNGTDM) -> float:
        if matrix.is_empty():
            return np.nan

        x = np.sum(matrix.pi.pi * matrix.pi.s)
        if x < 1.0E-6:
            return 1.0 / 1.0E-6
        else:
            return 1.0 / x


class FeatureNGTDMContrast(FeatureNGTDM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "NGTDM - contrast"
        self.abbr_name = "ngt_contrast"
        self.ibsi_id = "65HE"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixNGTDM) -> float:
        if matrix.is_empty():
            return np.nan

        if matrix.n_p > 1.0:
            return (
                    np.sum(matrix.pij.pi * matrix.pij.pj * (matrix.pij.i - matrix.pij.j) ** 2.0)
                    / (matrix.n_p * (matrix.n_p - 1.0)) * np.sum(matrix.pi.s) / matrix.n_voxels
            )
        else:
            return 0.0


class FeatureNGTDMBusyness(FeatureNGTDM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "NGTDM - busyness"
        self.abbr_name = "ngt_busyness"
        self.ibsi_id = "NQ30"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixNGTDM) -> float:
        if matrix.is_empty():
            return np.nan

        denominator = np.sum(np.abs(matrix.pij.i * matrix.pij.pi - matrix.pij.j * matrix.pij.pj))
        if matrix.n_p > 1.0 and denominator > 0.0:
            return np.sum(matrix.pi.pi * matrix.pi.s) / denominator
        else:
            return 0.0


class FeatureNGTDMComplexity(FeatureNGTDM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "NGTDM - complexity"
        self.abbr_name = "ngt_complexity"
        self.ibsi_id = "HDEZ"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixNGTDM) -> float:
        if matrix.is_empty():
            return np.nan

        return np.sum(
            np.abs(matrix.pij.i - matrix.pij.j)
            * (matrix.pij.pi * matrix.pij.si + matrix.pij.pj * matrix.pij.sj)
            / (matrix.pij.pi + matrix.pij.pj)
        ) / matrix.n_voxels


class FeatureNGTDMStrength(FeatureNGTDM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "NGTDM - strength"
        self.abbr_name = "ngt_strength"
        self.ibsi_id = "1X9X"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixNGTDM) -> float:
        if matrix.is_empty():
            return np.nan

        denominator = np.sum(matrix.pi.s)
        if denominator > 0.0:
            return np.sum((matrix.pij.pi + matrix.pij.pj) * (matrix.pij.i - matrix.pij.j) ** 2.0) / denominator
        else:
            return 0.0


def get_ngtdm_class_dict() -> dict[str, FeatureNGTDM]:
    class_dict = {
        "ngt_coarseness": FeatureNGTDMCoarseness,
        "ngt_contrast": FeatureNGTDMContrast,
        "ngt_busyness": FeatureNGTDMBusyness,
        "ngt_complexity": FeatureNGTDMComplexity,
        "ngt_strength": FeatureNGTDMStrength
    }

    return class_dict


def generate_ngtdm_features(
        settings: FeatureExtractionSettingsClass,
        features: None | list[str] = None
) -> Generator[FeatureNGTDM, None, None]:
    class_dict = get_ngtdm_class_dict()
    ngt_features = list(class_dict.keys())

    # Populate features if available.
    if features is None and settings.has_ngtdm_family():
        features = ngt_features

    # Terminate early if no features are set, and none are required.
    if features is None:
        return

    # Select only NGTDM-features, and return if none are present.
    features = [feature for feature in features if feature in ngt_features]
    if len(features) == 0:
        return

    # Features are parametrised by the choice of discretisation parameters and spatial methods.
    for discretisation_parameters in get_discretisation_parameters(
        settings=settings
    ):
        for spatial_method in settings.ngtdm_spatial_method:
            for feature in features:
                yield class_dict[feature](
                    spatial_method=spatial_method,
                    **discretisation_parameters
                )



import copy

from mirp._data_import.numpy_file import ImageNumpyFile, MaskNumpyFile
from mirp._data_import.generic_file_stack import ImageFileStack, MaskFileStack


class ImageNumpyFileStack(ImageFileStack):

    def __init__(
            self,
            image_file_objects: list[ImageNumpyFile],
            **kwargs
    ):
        super().__init__(
            image_file_objects=image_file_objects,
            **kwargs
        )

    def complete(self, remove_metadata=False, force=False):
        """
        Fills out missing attributes in an image stack. Numpy files are actually barebones, and compared to DICOM
        and many ITK formats lacks the origin. This forces us to determine stacking by the file name.
        :param remove_metadata: Whether metadata should be removed after completing information.
        :param force: Whether attributes are forced to update or not.
        :return: nothing, attributes are updated in place.
        """
        # Load metadata of every slice.
        self.load_metadata()

        self._complete_modality()
        self._complete_sample_name()

        # Sort by filename.
        self.sort_image_objects_by_file()

        image_object = copy.deepcopy(self.image_file_objects[0])
        image_object.complete(remove_metadata=False)

        if self.image_origin is None:
            self.image_origin = image_object.image_origin

        if self.image_spacing is None:
            self.image_spacing = image_object.image_spacing

        if self.image_orientation is None:
            self.image_orientation = image_object.image_orientation

        if self.image_dimension is None:
            image_dimension = image_object.image_dimension
            if len(image_dimension) == 3:
                self.image_dimension = tuple([
                    len(self.image_file_objects), image_object.image_dimension[1], image_object.image_dimension[2]
                ])
            elif len(image_dimension) == 2:
                self.image_dimension = tuple([
                    len(self.image_file_objects), image_object.image_dimension[0], image_object.image_dimension[1]
                ])
            else:
                self.image_dimension = tuple([len(self.image_file_objects), 1, image_object.image_dimension[0]])


class MaskNumpyFileStack(ImageNumpyFileStack, MaskFileStack):

    def __init__(
            self,
            image_file_objects: list[MaskNumpyFile],
            **kwargs
    ):

        super().__init__(
            image_file_objects=image_file_objects,
            **kwargs
        )

    def complete(self, remove_metadata=False, force=False):

        super().complete(remove_metadata=False, force=force)

        image_object = copy.deepcopy(self.image_file_objects[0])
        image_object.complete(remove_metadata=False)

        self.roi_name = image_object.roi_name



import warnings
from typing import Any
from xml.etree.ElementTree import Element

import numpy as np


def setting_def(
        arg_key: str,
        typing: str,
        to_list: bool = False,
        xml_key: None | str | list[str] = None,
        class_key: None | str = None,
        test: Any = None
) -> dict[str, Any]:

    if xml_key is None:
        xml_key = arg_key

    if class_key is None:
        class_key = arg_key

    if typing not in ["int", "float", "bool", "str", "path"]:
        raise TypeError(f"typing has an incorrect type: ", typing)

    return {
        "argument_key": arg_key,
        "xml_key": xml_key,
        "class_key": class_key,
        "typing": typing,
        "to_list": to_list,
        "test_value": test
    }


def find_branch(
        tree: None | Element,
        branch_name: str | list[str]
) -> None | Element:
    if tree is None:
        return None

    if isinstance(branch_name, str):
        branch_name = [branch_name]

    for current_branch_name in branch_name:
        branch = tree.find(current_branch_name)
        if isinstance(branch, Element):
            return branch

    return None


def update_settings_from_branch(
        kwargs: dict[str, Any],
        branch: None | Element,
        settings: list[dict[str, Any]]
):
    from mirp._data_import.utilities import flatten_list

    if branch is None:
        return

    # Iterate over all parameter in settings and update kwargs.
    for parameter in settings:
        is_parameter_set = False

        xml_key = parameter["xml_key"]
        if isinstance(xml_key, str):
            xml_key = [xml_key]

        # Iterate over all allowed xml tags for the current parameter, try to match the tag with one appearing in the
        # branch and then update kwargs with its contents.
        for current_xml_key in xml_key:
            for elem in branch.iter():
                if elem.tag == current_xml_key:
                    if elem.text is None:
                        continue

                    if parameter["to_list"]:
                        kwargs.update(dict([(
                            parameter["argument_key"],
                            str2list(elem.text, data_type=parameter["typing"])
                        )]))
                    else:
                        kwargs.update(dict([(
                            parameter["argument_key"],
                            str2type(elem.text, data_type=parameter["typing"])
                        )]))

                    is_parameter_set = True
                    break
            # If the parameter is set, stop iterating over the branch.
            if is_parameter_set:
                break

    # Check if any user-defined xml key is actually not a valid xml key.
    available_xml_keys = set(flatten_list([parameter["xml_key"] for parameter in settings]))
    observed_xml_keys = set([elem.tag for elem in branch.iter()])
    observed_xml_keys.remove(branch.tag)
    unknown_xml_keys = observed_xml_keys - available_xml_keys

    if len(unknown_xml_keys) > 0:
        raise ValueError(
            f"The following xml keys have been deprecated or misspelled, and are not used by mirp: "
            f"{', '.join([str(xml_key) for xml_key in unknown_xml_keys])}"
        )


def str2list(strx, data_type, default=None):
    """ Function for splitting strings read from the xml file """

    # Check if strx is none
    if strx is None and default is None:
        return None
    elif strx is None and type(default) in [list, tuple]:
        return default
    elif strx is None and not type(default) in [list, tuple]:
        return [default]

    # If strx is an element, read string
    if type(strx) is Element:
        strx = strx.text

    # Repeat check
    if strx is None and default is None:
        return None
    elif strx is None and type(default) in [list, tuple]:
        return default
    elif strx is None and not type(default) in [list, tuple]:
        return [default]

    contents = strx.split(",")
    content_list = []

    if (len(contents) == 1) and (contents[0] == ""):
        return content_list

    for i in np.arange(0, len(contents)):
        append_data = str2type(contents[i], data_type)

        # Check append data for None
        if append_data is None and type(default) in [list, tuple]:
            return default
        elif append_data is None and not type(default) in [list, tuple]:
            return [default]
        else:
            content_list.append(append_data)

    return content_list


def str2type(strx, data_type, default=None):
    # Check if strx is none
    if strx is None and default is None:
        return None
    elif strx is None:
        return default

    # If strx is an element, read string
    if isinstance(strx, Element):
        strx = strx.text

    # Strip white characters.
    if isinstance(strx, str):
        strx = strx.strip()

    # Test if the requested data type is not a string or path, but is empty
    if data_type not in ["str", "path"] and (strx == "" or strx is None):
        return default
    elif data_type in ["str", "path"] and (strx == "" or strx is None) and default is not None:
        return default

    # Casting of strings to different data types
    if data_type == "int":
        return int(strx)
    if data_type == "bool":
        if strx in ("true", "True", "TRUE", "T", "t", "1"):
            return True
        elif strx in ("false", "False", "FALSE", "F", "f", "0"):
            return False
    if data_type == "float":
        if strx in ("na", "nan", "NA", "NaN"):
            return np.nan
        elif strx in ("-inf", "-Inf", "-INF"):
            return -np.inf
        elif strx in ("inf", "Inf", "INF"):
            return np.inf
        else:
            return float(strx)
    if data_type == "str":
        return strx
    if data_type == "path":
        return strx


def read_node(
        tree: Element,
        node: str | list[str],
        deprecated_node: None | str | list[str] = None
):
    """
    :param tree: Tree element
    :param node: Name or list of names for each tree element.
    :param deprecated_node: Deprecated name.
    :return:
    """

    # Turn node into a list.
    if not isinstance(node, list):
        node = [node]

    # Throw deprecation warnings if necessary.
    if deprecated_node is not None:
        if not isinstance(deprecated_node, list):
            deprecated_node = [deprecated_node]

        for current_node in deprecated_node:
            if tree.find(current_node) is not None:
                warnings.warn(
                    f"The {current_node} has been deprecated. Use {', '.join(node)} instead.",
                    DeprecationWarning
                )

        # Append deprecated nodes to node.
        node += deprecated_node

    # Cycle over node, and return first instance without None.
    for current_node in node:
        node_contents = tree.find(current_node)
        if node_contents is not None:
            return node_contents

    return None



import copy
import warnings
from xml.etree.ElementTree import Element
from xml.etree import ElementTree as ElemTree

from mirp.settings.generic import SettingsClass
from mirp.settings.transformation_parameters import get_image_transformation_settings
from mirp.settings.feature_parameters import get_feature_extraction_settings
from mirp.settings.resegmentation_parameters import get_mask_resegmentation_settings
from mirp.settings.perturbation_parameters import get_perturbation_settings
from mirp.settings.image_processing_parameters import get_post_processing_settings
from mirp.settings.interpolation_parameters import get_image_interpolation_settings, get_mask_interpolation_settings
from mirp.settings.general_parameters import get_general_settings
from mirp.settings.utilities import update_settings_from_branch, find_branch


def create_settings_object(
    xml_tree: None | Element = None,
    **kwargs
) -> SettingsClass:
    kwargs = copy.deepcopy(kwargs)

    if isinstance(xml_tree, Element):
        # General settings
        update_settings_from_branch(
            kwargs=kwargs,
            branch=find_branch(xml_tree, "general"),
            settings=get_general_settings()
        )

        # Post-processing settings
        update_settings_from_branch(
            kwargs=kwargs,
            branch=find_branch(xml_tree, ["post_processing", "image_processing"]),
            settings=get_post_processing_settings()
        )

        # Image interpolation settings
        if xml_tree.find("img_interpolate") is not None and xml_tree.find("img_interpolate").find("new_non_iso_spacing") is not None:
            warnings.warn(
                f"The new_non_iso_spacing tag has been deprecated. Use the new_spacing tag instead.",
                DeprecationWarning
            )

        update_settings_from_branch(
            kwargs=kwargs,
            branch=find_branch(xml_tree, ["img_interpolate", "image_interpolation"]),
            settings=get_image_interpolation_settings()
        )

        # Mask interpolation settings
        update_settings_from_branch(
            kwargs=kwargs,
            branch=find_branch(xml_tree, ["roi_interpolate", "mask_interpolation"]),
            settings=get_mask_interpolation_settings()
        )

        # Perturbation settings
        update_settings_from_branch(
            kwargs=kwargs,
            branch=find_branch(xml_tree, ["vol_adapt", "image_perturbation"]),
            settings=get_perturbation_settings()
        )

        # Mask resegmentation settings
        update_settings_from_branch(
            kwargs=kwargs,
            branch=find_branch(xml_tree, ["roi_resegment", "mask_resegmentation"]),
            settings=get_mask_resegmentation_settings()
        )

        # Feature extraction settings
        if xml_tree.find("feature_extr") is not None and xml_tree.find("feature_extr").find("glcm_merge_method") is not None:
            warnings.warn(
                "The glcm_merge_method tag has been deprecated. Use the glcm_spatial_method tag instead. This takes"
                " the following values: `2d_average`, `2d_slice_merge`, '2.5d_direction_merge', '2.5d_volume_merge',"
                " '3d_average', and `3d_volume_merge`",
                DeprecationWarning
            )

        if xml_tree.find("feature_extr") is not None and xml_tree.find("feature_extr").find("glrlm_merge_method") is not None:
            warnings.warn(
                "The glrlm_merge_method tag has been deprecated. Use the glrlm_spatial_method tag instead. This "
                "takes the following values: `2d_average`, `2d_slice_merge`, '2.5d_direction_merge', "
                "'2.5d_volume_merge', '3d_average', and `3d_volume_merge`",
                DeprecationWarning
            )

        update_settings_from_branch(
            kwargs=kwargs,
            branch=find_branch(xml_tree, ["feature_extr", "feature_computation"]),
            settings=get_feature_extraction_settings()
        )

        # Image transformation settings
        if xml_tree.find("img_transform") is not None and xml_tree.find("img_transform").find("log_average") is not None:
            warnings.warn(
                "The log_average tag has been deprecated. Use the laplacian_of_gaussian_pooling_method tag "
                "instead with the value `mean` to emulate log_average=True.",
                DeprecationWarning
            )

        if xml_tree.find("img_transform") is not None and xml_tree.find("img_transform").find("riesz_steered") is not None:
            warnings.warn(
                "The riesz_steered tag has been deprecated. Steerable Riesz filter are now identified by the name "
                "of the filter kernel (filter_kernels parameter).",
                DeprecationWarning
            )

        update_settings_from_branch(
            kwargs=kwargs,
            branch=find_branch(xml_tree, ["img_transform", "image_transformation"]),
            settings=get_image_transformation_settings()
        )

        # Deep learning branch
        if xml_tree.find("deep_learning") is not None:
            warnings.warn(
                "deep_learning parameter branch has been deprecated. Parameters for image "
                "processing for deep learning can now be set directly using the deep_learning_preprocessing function.",
                DeprecationWarning
            )

    # Create settings class.
    settings = SettingsClass(**kwargs)

    return settings


def import_configuration_settings(
        compute_features: bool,
        path: None | str = None,
        **kwargs
) -> list[SettingsClass]:
    import os.path

    # Make a copy of the kwargs argument to avoid updating by reference.
    kwargs = copy.deepcopy(kwargs)

    # Prevent checking of feature parameters if features are not computed.
    if not compute_features:
        kwargs.update({
            "base_feature_families": "none",
            "response_map_feature_families": "none"
        })
    else:
        if "base_feature_families" not in kwargs:
            kwargs.update({"base_feature_families": "all"})
        if "response_map_feature_families" not in kwargs:
            kwargs.update({"response_map_feature_families": "statistics"})

    # Keywords only.
    if path is None:
        return [create_settings_object(xml_tree=None, **kwargs)]

    # Check path to xml file.
    if not os.path.exists(path):
        raise FileNotFoundError(f"The settings file could not be found at {path}.")

    # Load xml file.
    tree = ElemTree.parse(path)

    return [create_settings_object(xml_tree=branch, **kwargs) for branch in tree.getroot().findall("config")]




import copy

import numpy as np
import pandas as pd

from mirp._features.texture_matrix import Matrix


class MatrixNGLDM(Matrix):

    def __init__(
            self,
            distance: float,
            coarseness: float,
            **kwargs
    ):
        super().__init__(**kwargs)

        # Neighbourhood distance
        self.distance = int(distance)

        # Coarseness
        self.coarseness = coarseness

        # Placeholders for derivative values computed using set_values_from_matrix
        # Neighbourhood grey level dependence matrix
        self.sij: pd.DataFrame | None = None

        # Marginal sum over grey levels
        self.si: pd.DataFrame | None = None

        # Marginal sum over dependence counts
        self.sj: pd.DataFrame | None = None

        # Number of neighbourhoods
        self.n_s: int | None = None

    def compute(
            self,
            data: pd.DataFrame | None = None,
            image_dimension: tuple[int, int, int] | None = None,
            **kwargs
    ):
        # Check if data actually exists
        if data is None:
            return

        # Check if the roi contains any masked voxels. If this is not the case, don't construct the GLRLM.
        if not np.any(data.roi_int_mask):
            return

        if image_dimension is None:
            raise ValueError("image_dimension cannot be None, but may not have been provided in the calling function.")

        # Create local copies of the image table
        if self.spatial_method == "3d":
            data = copy.deepcopy(data)

            directions = list(self._generate_neighbour_direction(
                d=self.distance,
                metric="chebyshev",
                complete=True,
                dim3=True
            ))

        elif self.spatial_method in ["2d", "2.5d"]:
            data = copy.deepcopy(data[data.z == self.slice_id])
            data["index_id"] = np.arange(0, len(data))
            data["z"] = 0
            data = data.reset_index(drop=True)

            directions = list(self._generate_neighbour_direction(
                d=self.distance,
                metric="chebyshev",
                complete=True,
                dim3=False
            ))

        else:
            directions = None
            self._spatial_method_error()

        # Set grey level of voxels outside ROI to NaN
        data.loc[data.roi_int_mask == False, "g"] = np.nan

        # Set the number of voxels
        self.n_voxels = np.sum(data.roi_int_mask.values)

        # Initialise sum of grey levels and number of neighbours.
        data["occur"] = 0

        for direction in directions:
            # Determine potential valid neighbours.
            data["to_index"] = self.coord_to_index(
                x=data.x.values + direction[2],
                y=data.y.values + direction[1],
                z=data.z.values + direction[0],
                dims=image_dimension
            )

            # Get grey level value from transitions
            data["to_g"] = self._lookup_intensity(
                x=data.g.values,
                index=data.to_index.values
            )

            # Determine which voxels have valid neighbours
            valid_index = np.isfinite(data.to_g)

            # Determine co-occurrence within the coarseness
            data.loc[valid_index, "occur"] += ((np.abs(data.to_g - data.g)[valid_index]) <= self.coarseness) * 1

        # Work with voxels within the intensity roi
        data = data[data.roi_int_mask]

        # Drop superfluous columns
        data = data.drop(columns=["index_id", "x", "y", "z", "to_index", "to_g", "roi_int_mask"])

        # Sum s over voxels
        matrix = data.groupby(by=["g", "occur"]).size().reset_index(name="n")
        matrix.columns = ["i", "j", "sij"]

        # Add one to dependency count as features are not defined for k=0
        matrix.j += 1.0

        # Add matrix to object
        self.matrix = matrix

    def set_values_from_matrix(self, **kwargs):
        if self.is_empty():
            return

        # Neighbourhood grey level dependence matrix
        self.sij = copy.deepcopy(self.matrix)

        # Marginal sum over grey levels
        self.si = self.sij.groupby(by="i")["sij"].sum().reset_index().rename(columns={"sij": "si"})

        # Marginal sum over dependence counts
        self.sj = self.sij.groupby(by="j")["sij"].sum().reset_index().rename(columns={"sij": "sj"})

        # Number of neighbourhoods
        self.n_s = np.sum(self.sij.sij)

    @staticmethod
    def _get_grouping_columns():
        return ["i", "j"]



import copy
import sys
import warnings

import numpy as np

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class BaseImage:

    def __init__(
            self,
            sample_name: None | str = None,
            image_modality: None | str = None,
            image_origin: None | tuple[float, ...] = None,
            image_orientation: None | np.ndarray = None,
            image_spacing: None | tuple[float, ...] = None,
            image_dimensions: None | tuple[int, ...] = None,
            separate_slices: None | bool = None,
            metadata: None | dict[str, str] = None,
            **kwargs
    ):
        # Make cooperative
        super().__init__()

        # Set modality.
        if isinstance(image_modality, str):
            image_modality = image_modality.lower()
        self.modality = image_modality

        # Set affine-transformation related values.
        self.image_origin = copy.deepcopy(image_origin)
        self.image_orientation = copy.deepcopy(image_orientation)
        self.image_spacing = copy.deepcopy(image_spacing)
        self.image_dimension = copy.deepcopy(image_dimensions)

        # Set sample name.
        self.sample_name = sample_name

        # Initialise associated masks, which is required for passing image
        self.image_data = None
        self.associated_masks = None

        # Set metadata. Entries with empty values (None or "") are removed from the metadata dict to avoid
        # polluting the dictionary with unset entries.
        if isinstance(metadata, dict) and len(metadata) > 0:
            metadata = copy.deepcopy(metadata)
            remove_keys = [key for key, value in metadata.items() if value is None or value == ""]
            for current_key in remove_keys:
                metadata.pop(current_key, None)

            if len(metadata) > 0:
                self.object_metadata = metadata
            else:
                self.object_metadata = dict()
        else:
            self.object_metadata = dict()

        # Determines whether slices in the stack should be treated separately.
        self.separate_slices = separate_slices

    def copy(self, drop_image=False) -> Self:
        image = copy.deepcopy(self)

        if drop_image:
            image.drop_image()

        return image

    def drop_image(self):
        self.image_data = None

    def is_isotropic(self) -> bool:
        if self.separate_slices:
            spacing = np.array(self.image_spacing)[[1, 2]]
        else:
            spacing = np.array(self.image_spacing)

        return np.all(spacing == spacing[0])

    def set_modality(self, modality: None | str):
        from mirp._data_import.utilities import supported_image_modalities
        if modality is None:
            return

        if not isinstance(modality, str):
            raise ValueError(f"modality is expected to be a character string. Found: {modality}")

        if modality == "generic":
            raise ValueError(f"modality cannot be 'generic'")

        modality = supported_image_modalities(modality)
        if self.modality is None or self.modality == "generic":
            self.modality = modality[0]

    @staticmethod
    def get_dir_path():
        # BaseImage does not have an associated directory path.
        return None

    @staticmethod
    def get_file_name():
        # BaseImage also has no associated file name.
        return None

    def get_image_origin(self, as_str=False):
        if not as_str:
            return self.image_origin

        if self.image_origin is None:
            return "unset_image_origin"

        return str(self.image_origin)

    def get_image_orientation(self, as_str=False):
        if not as_str:
            return self.image_orientation

        if self.image_orientation is None:
            return "unset_image_orientation"

        return str(np.ravel(self.image_orientation))

    def get_image_spacing(self, as_str=False):
        if not as_str:
            return self.image_spacing

        if self.image_spacing is None:
            return "unset_image_spacing"

        return str(self.image_spacing)

    def get_image_dimension(self, as_str=False):
        if not as_str:
            return self.image_dimension

        if self.image_dimension is None:
            return "unset_image_dimension"

        return str(self.image_dimension)

    def remove_metadata(self, force=False):
        if force:
            self.object_metadata = dict()

    def associate_with_mask(
            self,
            mask_list,
            association_strategy: None | set[str] = None
    ):
        if mask_list is None or len(mask_list) == 0 or association_strategy is None:
            return

        # Match on sample name.
        if "sample_name" in association_strategy and self.sample_name is not None:
            matching_mask_list = [
                mask_file for mask_file in mask_list
                if self.sample_name == mask_file.sample_name
            ]

            if len(matching_mask_list) > 0:
                self.associated_masks = matching_mask_list
                return

        return

    def on_file_system(self):
        # The BaseImage object is by its nature not on the file system.
        return False

    def check_associated_masks(self):
        if self.associated_masks is None:
            return

        for mask in self.associated_masks:
            self._check_associated_mask_image_data(mask=mask)

    def _check_associated_mask_image_data(self, mask):
        """
        Check whether image and associated mask plausibly share the same frame of reference. This method is only
        used during import of BaseImage and BaseMask.
        """

        problem_list = []
        # Mismatch in grid dimension
        if not np.array_equal(self.get_image_dimension(), mask.get_image_dimension()):
                problem_list += [
                    f"different dimensions: \n\t\timage: {self.get_image_dimension()}\n\t\tmask: {mask.get_image_dimension()}"
                ]

        # Mismatch in origin
        if not np.allclose(self.get_image_origin(), mask.get_image_origin()):
            problem_list += [
                f"different origin: \n\t\timage: {self.get_image_origin()}\n\t\tmask: {mask.get_image_origin()}"
            ]

        # Mismatch in spacing
        if not np.allclose(self.get_image_spacing(), mask.get_image_spacing()):
            problem_list += [
                f"different spacing: \n\t\timage: {self.get_image_spacing()}\n\t\tmask: {mask.get_image_spacing()}"
            ]

        # Mismatch in orientation
        if not np.allclose(self.get_image_orientation(), mask.get_image_orientation()):
            problem_list += [
                f"different orientation: \n\t\timage: {np.ravel(self.get_image_orientation())}\n\t\tmask: "
                f"{np.ravel(mask.get_image_orientation())}"
            ]

        if len(problem_list) > 0:
            warnings.warn(
                f"Image and mask may not have the same frame of "
                f"reference. Please check if segmentation masks are placed correctly:\n\t" + "\n\t".join(problem_list),
                UserWarning
            )

    def to_object(self, **kwargs):
        image = self.copy()

        # Drop associated masks.
        image.associated_masks = None

        return image

    def promote(self, **kwargs):
        return self

    def world_coordinates(self):

        # Create grid.
        voxel_coordinate_z, voxel_coordinate_y, voxel_coordinate_x = np.mgrid[
            :self.image_dimension[0],
            :self.image_dimension[1],
            :self.image_dimension[2]
        ]

        # Set voxel coordinates.
        voxel_coordinates = np.array([
            voxel_coordinate_z.flatten().astype(float),
            voxel_coordinate_y.flatten().astype(float),
            voxel_coordinate_x.flatten().astype(float)])

        # Convert world coordinates.
        return self.to_world_coordinates(x=voxel_coordinates)

    def to_world_coordinates(
            self,
            x: np.ndarray,
            origin: None | np.ndarray = None,
            orientation: None | np.ndarray = None,
            spacing: None | np.ndarray = None,
            trim_result: bool = True
    ) -> np.ndarray:
        """
        This mirrors ImageClass.to_world_coordinates
        :param x:
        :param origin:
        :param orientation:
        :param spacing:
        :param trim_result:
        :return:
        """

        # Add empty dimension if the coordinates are unset.
        if len(x.shape) == 1:
            squeeze_on_exit = True
            x = x[:, np.newaxis]

        else:
            squeeze_on_exit = False

        # Setup voxel coordinates
        voxel_coordinates = np.ones([4, x.shape[1]])
        voxel_coordinates[0:3, :] = x

        # Define affine matrix.
        affine_matrix = self.get_affine_matrix(
            origin=origin,
            orientation=orientation,
            spacing=spacing,
            inverse=False
        )

        # World coordinates
        world_coordinates = np.matmul(affine_matrix, voxel_coordinates)

        if trim_result:
            world_coordinates = world_coordinates[0:3, :]

        if squeeze_on_exit:
            world_coordinates = np.squeeze(world_coordinates, axis=1)

        return world_coordinates

    def to_voxel_coordinates(
            self,
            x: np.ndarray,
            origin: None | np.ndarray = None,
            orientation: None | np.ndarray = None,
            spacing: None | np.ndarray = None,
            trim_result: bool = True
    ) -> np.ndarray:
        """
        This mirrors ImageClass.to_voxel_coordinates
        :param x:
        :param origin:
        :param orientation:
        :param spacing:
        :param trim_result:
        :return:
        """

        # Add empty dimension if the coordinates are unset.
        if len(x.shape) == 1:
            squeeze_on_exit = True
            x = x[:, np.newaxis]

        else:
            squeeze_on_exit = False

        # Setup voxel coordinates
        world_coordinates = np.ones([4, x.shape[1]])
        world_coordinates[0:3, :] = x

        # Define inverse affine matrix.
        inverse_affine_matrix = self.get_affine_matrix(
            origin=origin,
            orientation=orientation,
            spacing=spacing,
            inverse=True
        )

        # Convert from world to voxel coordinates
        voxel_coordinates = np.matmul(inverse_affine_matrix, world_coordinates)

        if trim_result:
            voxel_coordinates = voxel_coordinates[0:3, :]

        if squeeze_on_exit:
            voxel_coordinates = np.squeeze(voxel_coordinates, axis=1)

        return voxel_coordinates

    def get_affine_matrix(
            self,
            origin: None | np.ndarray = None,
            orientation: None | np.ndarray = None,
            spacing: None | np.ndarray = None,
            inverse: bool = False
    ) -> np.ndarray:
        """
        This mirrors ImageClass.get_affine_matrix.

        An affine matrix can be used to convert between local (voxel) and world coordinates:
          W = A_origin * A_orientation * A_spacing * X
            = A_affine * X

        Here we use the fact that all right-hand matrices except X are square matrices with the same dimensions,
        and thus are associative. Here we compute the result of A_origin * A_orientation * A_spacing.

        Parameters
        ----------
        origin: np.ndarray, optional, default: None
            Origin of the voxel grid in world coordinates. If None, the origin of the image is used.

        orientation: np.ndarray, optional, default: None
            Orientation of the voxel grid in the world coordinate frame. If None, the orientation of
            the image is used.

        spacing: np.ndarray, optional, default: None
            Spacing of voxels in world units. If None, the spacing of the image is used.

        inverse: bool, default: False
            Return inverse of the affine matrix

        Returns
        -------
        np.ndarray
            Affine matrix or inverse affine matrix (if `inverse=True`)
        """

        # Matrix multiplication of orientation and spacing (scale) matrices (A_orientation * A_spacing).
        matrix = np.matmul(
            self._get_orientation_matrix(orientation=orientation), self._get_spacing_matrix(spacing=spacing)
        )

        # Matrix multiplication of the origin matrix and the previous result (A_origin * (A_orientation * A_spacing)).
        matrix = np.matmul(self._get_origin_matrix(origin=origin), matrix)

        # Compute inverse affine matrix for translating from world coordinates to voxel coordinates.
        if inverse:
            matrix = np.linalg.inv(matrix)

        return matrix

    def _get_orientation_matrix(
            self,
            size: int = 4,
            orientation: None | np.ndarray = None
    ) -> np.ndarray:
        """
        This mirrors ImageClass._get_orientation_matrix
        :param size:
        :param orientation:
        :return:
        """

        if size not in [3, 4]:
            raise ValueError(f"The size argument should be 3 or 4. Found: {size}")

        if orientation is None:
            orientation = self.image_orientation

        # Create identity matrix and insert orientation matrix.
        orientation_matrix = np.identity(n=size, dtype=float)
        orientation_matrix[0:3, 0:3] = orientation

        return orientation_matrix

    def _get_spacing_matrix(
            self,
            size: int = 4,
            spacing: None | np.ndarray = None,
            inverse: bool = False
    ) -> np.ndarray:
        """
        This mirrors ImageClass._get_spacing_matrix
        :param size:
        :param spacing:
        :param inverse:
        :return:
        """

        if size not in [3, 4]:
            raise ValueError(f"The size argument should be 3 or 4. Found: {size}")

        if spacing is None:
            spacing = np.array(self.image_spacing)

        if size == 4:
            spacing = np.append(spacing, 1.0)

        if inverse:
            spacing = 1.0 / spacing

        # Create identity matrix.
        spacing_matrix = np.identity(n=size, dtype=float)

        # Fill diagonal of the identity matrix with spacing.
        np.fill_diagonal(spacing_matrix, spacing)

        return spacing_matrix

    def _get_origin_matrix(
            self,
            origin: None | np.ndarray = None,
            inverse: bool = False
    ) -> np.ndarray:
        """
        This mirrors ImageClass._get_origin_matrix
        :param origin:
        :param inverse:
        :return:
        """

        if origin is None:
            origin = np.array(self.image_origin)

        if inverse:
            origin = -1.0 * origin

        origin_matrix = np.identity(n=4, dtype=float)
        origin_matrix[0:3, 3] = origin

        return origin_matrix



from functools import lru_cache
from typing import Generator

import numpy as np

from mirp._features.dzm_matrix import MatrixDZM
from mirp._images.generic_image import GenericImage
from mirp._masks.base_mask import BaseMask
from mirp._features.histogram import get_discretisation_parameters
from mirp._features.texture_features import FeatureTexture
from mirp.settings.feature_parameters import FeatureExtractionSettingsClass


class FeatureDZM(FeatureTexture):

    def __init__(
            self,
            **kwargs
    ):
        super().__init__(**kwargs)

        # Perform close crop for DZM.
        self.cropping_distance = 0.0

    def get_matrix(
            self,
            image: GenericImage,
            mask: BaseMask,
    ) -> list[MatrixDZM]:
        # First discretise the image.
        image, mask = self.discretise_image(
            image=image,
            mask=mask,
            discretisation_method=self.discretisation_method,
            bin_width=self.bin_width,
            bin_number=self.bin_number,
            cropping_distance=self.cropping_distance
        )

        # Then get matrix or matrices
        matrix_list = self._get_matrix(
            image=image,
            mask=mask,
            spatial_method=self.spatial_method
        )

        return matrix_list

    @staticmethod
    @lru_cache(maxsize=1)
    def _get_matrix(
            image: GenericImage,
            mask: BaseMask,
            spatial_method: str
    ) -> list[MatrixDZM]:
        # Extract data from the mask and image.
        data = mask.as_pandas_dataframe(
            image=image,
            intensity_mask=True,
            distance_map=True,
            by_slice=spatial_method in ["2d", "2.5d"]
        )

        # Instantiate a helper copy of the current class to be able to use class methods without tying the cache to the
        # instance of the original object from which this method is called.
        matrix_instance = MatrixDZM(
            spatial_method=spatial_method
        )

        # Compute the required matrices.
        matrix_list = list(matrix_instance.generate(prototype=MatrixDZM, n_slices=image.image_dimension[0]))
        for matrix in matrix_list:
            matrix.compute(data=data, image=image, mask=mask)

        # Merge according to the spatial method.
        matrix_list = matrix_instance.merge(matrix_list, prototype=MatrixDZM)

        # Compute additional values from the individual matrices.
        for matrix in matrix_list:
            matrix.set_values_from_matrix()

        return matrix_list

    def clear_local_cache(self, other):
        if not isinstance(other, FeatureDZM):
            self._get_matrix.cache_clear()

    def clear_cache(self):
        super().clear_cache()
        self._get_matrix.cache_clear()

    def compute(self, image: GenericImage, mask: BaseMask):
        # Skip processing if input image and/or roi are missing
        if image is None or mask is None:
            return

        # Check if data actually exists
        if image.is_empty() or mask.roi_intensity.is_empty_mask():
            return

        # Compute or retrieve matrices from cache.
        matrices = self.get_matrix(image=image, mask=mask)

        # Compute feature value from matrices, and average over matrices.
        values = [self._compute(matrix=matrix) for matrix in matrices]
        if np.all(np.isnan(values)):
            self.value = np.nan
        else:
            self.value = np.nanmean(values)

    @staticmethod
    def _compute(matrix: MatrixDZM):
        raise NotImplementedError("Implement _compute for feature-specific computation.")

    def create_table_name(self):
        table_elements = (
                self._get_base_table_name_element()
                + self._get_spatial_table_name_element()
                + self._get_discretisation_table_name_element()
        )
        self.table_name = "_".join(table_elements)


class FeatureDZMSmallDistanceEmphasis(FeatureDZM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "DZM - small distance emphasis"
        self.abbr_name = "dzm_sde"
        self.ibsi_id = "0GBI"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixDZM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.dj.dj / matrix.dj.j ** 2.0) / matrix.n_s


class FeatureDZMLargeDistanceEmphasis(FeatureDZM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "DZM - large distance emphasis"
        self.abbr_name = "dzm_lde"
        self.ibsi_id = "MB4I"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixDZM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.dj.dj * matrix.dj.j ** 2.0) / matrix.n_s


class FeatureDZMLowGreyLevelZoneEmphasis(FeatureDZM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "DZM - low grey level zone emphasis"
        self.abbr_name = "dzm_lgze"
        self.ibsi_id = "S1RA"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixDZM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.di.di / matrix.di.i ** 2.0) / matrix.n_s


class FeatureDZMHighGreyLevelZoneEmphasis(FeatureDZM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "DZM - high grey level zone emphasis"
        self.abbr_name = "dzm_hgze"
        self.ibsi_id = "K26C"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixDZM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.di.di * matrix.di.i ** 2.0) / matrix.n_s


class FeatureDZMSmallDistanceLowGreyLevelEmphasis(FeatureDZM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "DZM - small distance low grey level emphasis"
        self.abbr_name = "dzm_sdlge"
        self.ibsi_id = "RUVG"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixDZM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.dij.dij / (matrix.dij.i * matrix.dij.j) ** 2.0) / matrix.n_s


class FeatureDZMSmallDistanceHighGreyLevelEmphasis(FeatureDZM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "DZM - small distance high grey level emphasis"
        self.abbr_name = "dzm_sdhge"
        self.ibsi_id = "DKNJ"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixDZM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.dij.dij * matrix.dij.i ** 2.0 / matrix.dij.j ** 2.0) / matrix.n_s


class FeatureDZMLargeDistanceLowGreyLevelEmphasis(FeatureDZM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "DZM - large distance low grey level emphasis"
        self.abbr_name = "dzm_ldlge"
        self.ibsi_id = "A7WM"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixDZM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.dij.dij * matrix.dij.j ** 2.0 / matrix.dij.i ** 2.0) / matrix.n_s


class FeatureDZMLargeDistanceHighGreyLevelEmphasis(FeatureDZM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "DZM - large distance high grey level emphasis"
        self.abbr_name = "dzm_ldhge"
        self.ibsi_id = "KLTH"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixDZM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.dij.dij * matrix.dij.i ** 2.0 * matrix.dij.j ** 2.0) / matrix.n_s


class FeatureDZMGreyLevelNonUniformity(FeatureDZM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "DZM - grey level non-uniformity"
        self.abbr_name = "dzm_glnu"
        self.ibsi_id = "VFT7"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixDZM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.di.di ** 2.0) / matrix.n_s


class FeatureDZMNormalisedGreyLevelNonUniformity(FeatureDZM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "DZM - normalised grey level non-uniformity"
        self.abbr_name = "dzm_glnu_norm"
        self.ibsi_id = "7HP3"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixDZM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.di.di ** 2.0) / matrix.n_s ** 2.0


class FeatureDZMZoneDistanceNonUniformity(FeatureDZM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "DZM - zone distance non-uniformity"
        self.abbr_name = "dzm_zdnu"
        self.ibsi_id = "V294"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixDZM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.dj.dj ** 2.0) / matrix.n_s


class FeatureDZMNormalisedZoneDistanceNonUniformity(FeatureDZM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "DZM - normalised zone distance non-uniformity"
        self.abbr_name = "dzm_zdnu_norm"
        self.ibsi_id = "IATH"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixDZM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.dj.dj ** 2.0) / matrix.n_s ** 2.0


class FeatureDZMZonePercentage(FeatureDZM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "DZM - zone percentage"
        self.abbr_name = "dzm_z_perc"
        self.ibsi_id = "VIWW"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixDZM) -> float:
        if matrix.is_empty():
            return np.nan
        return matrix.n_s / matrix.n_voxels


class FeatureDZMGreyLevelVariance(FeatureDZM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "DZM - grey level variance"
        self.abbr_name = "dzm_gl_var"
        self.ibsi_id = "QK93"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixDZM) -> float:
        if matrix.is_empty():
            return np.nan
        mu = np.sum(matrix.dij.dij * matrix.dij.i) / matrix.n_s
        return np.sum((matrix.dij.i - mu) ** 2.0 * matrix.dij.dij) / matrix.n_s


class FeatureDZMZoneDistanceVariance(FeatureDZM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "DZM - zone distance variance"
        self.abbr_name = "dzm_zd_var"
        self.ibsi_id = "7WT1"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixDZM) -> float:
        if matrix.is_empty():
            return np.nan
        mu = np.sum(matrix.dij.dij * matrix.dij.j) / matrix.n_s
        return np.sum((matrix.dij.j - mu) ** 2.0 * matrix.dij.dij) / matrix.n_s


class FeatureDZMZoneDistanceEntropy(FeatureDZM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "DZM - zone distance entropy"
        self.abbr_name = "dzm_zd_entr"
        self.ibsi_id = "GBDU"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixDZM) -> float:
        if matrix.is_empty():
            return np.nan
        return -np.sum(matrix.dij.dij * np.log2(matrix.dij.dij / matrix.n_s)) / matrix.n_s


def get_dzm_class_dict() -> dict[str, FeatureDZM]:
    class_dict = {
        "dzm_sde": FeatureDZMSmallDistanceEmphasis,
        "dzm_lde": FeatureDZMLargeDistanceEmphasis,
        "dzm_lgze": FeatureDZMLowGreyLevelZoneEmphasis,
        "dzm_hgze": FeatureDZMHighGreyLevelZoneEmphasis,
        "dzm_sdlge": FeatureDZMSmallDistanceLowGreyLevelEmphasis,
        "dzm_sdhge": FeatureDZMSmallDistanceHighGreyLevelEmphasis,
        "dzm_ldlge": FeatureDZMLargeDistanceLowGreyLevelEmphasis,
        "dzm_ldhge": FeatureDZMLargeDistanceHighGreyLevelEmphasis,
        "dzm_glnu": FeatureDZMGreyLevelNonUniformity,
        "dzm_glnu_norm": FeatureDZMNormalisedGreyLevelNonUniformity,
        "dzm_zdnu": FeatureDZMZoneDistanceNonUniformity,
        "dzm_zdnu_norm": FeatureDZMNormalisedZoneDistanceNonUniformity,
        "dzm_z_perc": FeatureDZMZonePercentage,
        "dzm_gl_var": FeatureDZMGreyLevelVariance,
        "dzm_zd_var": FeatureDZMZoneDistanceVariance,
        "dzm_zd_entr": FeatureDZMZoneDistanceEntropy
    }

    return class_dict


def generate_dzm_features(
        settings: FeatureExtractionSettingsClass,
        features: None | list[str] = None
) -> Generator[FeatureDZM, None, None]:
    class_dict = get_dzm_class_dict()
    dzm_features = list(class_dict.keys())

    # Populate features if available.
    if features is None and settings.has_gldzm_family():
        features = dzm_features

    # Terminate early if no features are set, and none are required.
    if features is None:
        return

    # Select only DZM-features, and return if none are present.
    features = [feature for feature in features if feature in dzm_features]
    if len(features) == 0:
        return

    # Features are parametrised by the choice of discretisation parameters and spatial methods.
    for discretisation_parameters in get_discretisation_parameters(
        settings=settings
    ):
        for spatial_method in settings.gldzm_spatial_method:
            for feature in features:
                yield class_dict[feature](
                    spatial_method=spatial_method,
                    **discretisation_parameters
                )



from typing import Generator

import numpy as np
import copy

from mirp._images.generic_image import GenericImage
from mirp._images.transformed_image import LocalBinaryPatternImage
from mirp.settings.generic import SettingsClass
from mirp._imagefilters.generic import GenericFilter


class LocalBinaryPatternFilter(GenericFilter):

    def __init__(self, image: GenericImage, settings: SettingsClass, name: str):

        super().__init__(image=image, settings=settings, name=name)

        self.ibsi_compliant = False

        self.separate_slices = settings.img_transform.lbp_separate_slices
        self.lbp_method = settings.img_transform.lbp_method
        self.d = settings.img_transform.lbp_distance

    def generate_object(self):
        # Generator for transformation objects.
        lbp_method = copy.deepcopy(self.lbp_method)
        if not isinstance(lbp_method, list):
            lbp_method = [lbp_method]

        separate_slices = copy.deepcopy(self.separate_slices)
        if not isinstance(separate_slices, list):
            separate_slices = [separate_slices]

        distance = copy.deepcopy(self.d)
        if not isinstance(distance, list):
            distance = [distance]

        # Iterate over options to yield filter objects with specific settings. A copy of the parent object is made to
        # avoid updating by reference.
        for current_separate_slices in separate_slices:
            for current_lbp_method in lbp_method:
                for current_distance in distance:
                    filter_object = copy.deepcopy(self)
                    filter_object.separate_slices = current_separate_slices
                    filter_object.lbp_method = current_lbp_method
                    filter_object.d = current_distance

                    yield filter_object

    def transform(self, image: GenericImage) -> LocalBinaryPatternImage:
        # Create placeholder Laplacian-of-Gaussian response map.
        response_map = LocalBinaryPatternImage(
            image_data=None,
            separate_slices=self.separate_slices,
            distance=self.d,
            lbp_method = self.lbp_method,
            template=image
        )
        response_map.ibsi_compliant = self.ibsi_compliant and image.ibsi_compliant

        if image.is_empty():
            return response_map

        response_map.set_voxel_grid(
            voxel_grid=self.transform_grid(voxel_grid=image.get_voxel_grid())
        )

        return response_map

    def transform_grid(
            self,
            voxel_grid: np.ndarray
    ):
        # Voxel grid as a contiguous flattened array.
        dims = voxel_grid.shape
        voxel_original = np.ravel(voxel_grid)

        # # Get directions corresponding to distance (d) and separate_slices. For each direction, determine the
        # position of voxels with a valid neighbour (i.e. not out-of-volume) and the
        # position of their neighbour.
        neighbour_vectors = list(self._generate_neighbour_direction())
        weights = 2 ** np.arange(len(neighbour_vectors))

        # Initialise response map as a flattened array.
        lbp = np.zeros((len(neighbour_vectors), len(voxel_original)), dtype = bool)
        for ii, neighbour_vector in enumerate(neighbour_vectors):
            mask, voxel_neighbour = self._lookup_neighbour_voxel_value(
                voxels=voxel_original,
                dims=dims,
                lookup_vector=neighbour_vector
            )
            lbp[ii, mask] = voxel_neighbour - voxel_original[mask] >= 0.0

        if self.lbp_method == "default":
            voxel_response = np.sum(np.multiply(lbp, weights[:, np.newaxis]), axis = 0)

        elif self.lbp_method == "variance":
            voxel_response = np.var(lbp, axis = 0)

        elif self.lbp_method == "rotation_invariant":
            voxel_response = np.sum(np.multiply(lbp, weights[:, np.newaxis]), axis=0)

            for ii in np.arange(1, len(neighbour_vectors)):
                lbp = np.roll(lbp, 1, axis = 0)
                voxel_response = np.min(np.vstack((
                    voxel_response,
                    np.sum(np.multiply(lbp, weights[:, np.newaxis]), axis=0)
                )), axis =0)

        else:
            raise ValueError(f"Unknown method: {self.lbp_method}")

        return np.reshape(voxel_response, shape = dims)

    @staticmethod
    def _coord_to_index(z, y, x, dims):
        # Translate coordinates to indices
        index = x + y * dims[2] + z * dims[2] * dims[1]

        # Mark invalid transitions
        index[np.logical_or(x < 0, x >= dims[2])] = -99999
        index[np.logical_or(y < 0, y >= dims[1])] = -99999
        index[np.logical_or(z < 0, z >= dims[0])] = -99999

        return index

    @staticmethod
    def _index_to_coord(index, dims):
        z = index // (dims[2] * dims[1])
        index -= z * (dims[2] * dims[1])
        y = index // (dims[2])
        x = index - y * dims[2]

        return z, y, x

    def _lookup_neighbour_voxel_value(self, voxels, dims, lookup_vector):
        z, y, x = self._index_to_coord(index=np.arange(len(voxels)), dims=dims)
        neighbour_index = self._coord_to_index(
            z = z + lookup_vector[0],
            y = y + lookup_vector[1],
            x = x + lookup_vector[2],
            dims = dims
        )

        mask = neighbour_index > 0
        return mask, voxels[neighbour_index[mask]]

    def _generate_neighbour_direction(self) -> Generator[tuple[int, ...], None, None]:
        from mirp._features.utilities import rep

        if self.separate_slices:
            m = 8
            nbrs = np.array([
                np.zeros(m, int),
                np.round(self.d * np.sin(2 * np.pi * np.arange(m, dtype=float) / m)),
                np.round(self.d * np.cos(2 * np.pi * np.arange(m, dtype=float) / m))
            ], dtype = int)

            # Remove duplicates
            _, indices = np.unique(nbrs, return_index=True, axis=1)
            nbrs = nbrs[:, indices.sort()].squeeze()

            # Compute distance to eliminate
            neighbour_distance = np.sqrt(np.sum(np.multiply(nbrs, nbrs), axis=0))
            index = neighbour_distance > 0.0

            for ii, flag in enumerate(index):
                if flag:
                    yield tuple(nbrs[:, ii].flatten())

        else:
            # Base transition vector
            trans = np.arange(start=-np.ceil(self.d + 1.0), stop=np.ceil(self.d + 1.0) + 1)
            n = np.size(trans)

            # Build transition array [z,y,x]
            nbrs = np.array([
                rep(x=trans, each=n * n, times=1),
                rep(x=trans, each=n, times=n, use_inversion=True),
                rep(x=trans, each=1, times=n * n, use_inversion=True)
            ], dtype=int)

            # Filter neighbours based on distance. That is, all voxels that fall within distance d and d-1.0 (a single
            # rim of voxels), and excluding the central voxel.
            neighbour_distance = np.sqrt(np.sum(np.multiply(nbrs, nbrs), axis = 0))
            index = np.logical_and(neighbour_distance <= self.d, neighbour_distance > self.d - 1.0)
            index = np.logical_and(index, neighbour_distance > 0.0)

            for ii, flag in enumerate(index):
                if flag:
                    yield tuple(nbrs[:, ii].flatten())



import numpy as np

from mirp._image_processing.utilities import standard_image_process_checks
from mirp._images.generic_image import GenericImage
from mirp._images.mask_image import MaskImage
from mirp._masks.base_mask import BaseMask


def crop(
        image: GenericImage,
        masks: BaseMask | MaskImage | list[BaseMask],
        boundary: float = 0.0,
        xy_only: bool = False,
        z_only: bool = False,
        in_place: bool = False
) -> tuple[GenericImage, None | BaseMask | MaskImage | list[BaseMask]]:
    """ The function is used to slice a subsection of the image so that further processing is facilitated in terms of
     memory and computational requirements. """

    image, masks, return_list = standard_image_process_checks(image=image, masks=masks)
    if return_list is None:
        return image, None

    bounds_z: None | list[int] = None
    bounds_y: None | list[int] = None
    bounds_x: None | list[int] = None

    for mask in masks:
        current_bounds_z, current_bounds_y, current_bounds_x = mask.get_bounding_box()
        if current_bounds_z is None or current_bounds_y is None or current_bounds_x is None:
            continue

        if bounds_z is None:
            bounds_z = list(current_bounds_z)
        else:
            bounds_z = [min(bounds_z[0], current_bounds_z[0]), max(bounds_z[1], current_bounds_z[1])]

        if bounds_y is None:
            bounds_y = list(current_bounds_y)
        else:
            bounds_y = [min(bounds_y[0], current_bounds_y[0]), max(bounds_y[1], current_bounds_y[1])]

        if bounds_x is None:
            bounds_x = list(current_bounds_x)
        else:
            bounds_x = [min(bounds_x[0], current_bounds_x[0]), max(bounds_x[1], current_bounds_x[1])]

    # Check if bounds were determined.
    if bounds_x is None or bounds_y is None or bounds_z is None:
        if return_list:
            return image, masks
        else:
            return image, masks[0]

    # Compute boundary and add to bounding box.
    boundary = np.ceil(boundary / np.array(image.image_spacing)).astype(int)

    if not image.separate_slices:
        bounds_z = [bounds_z[0] - boundary[0], bounds_z[1] + boundary[0]]
    bounds_y = [bounds_y[0] - boundary[1], bounds_y[1] + boundary[1]]
    bounds_x = [bounds_x[0] - boundary[2], bounds_x[1] + boundary[2]]

    if not in_place:
        image = image.copy()
        masks = [mask.copy() for mask in masks]

    # Crop images and masks.
    image.crop(
        ind_ext_z=bounds_z,
        ind_ext_y=bounds_y,
        ind_ext_x=bounds_x,
        xy_only=xy_only,
        z_only=z_only
    )

    for mask in masks:
        mask.crop(
            ind_ext_z=bounds_z,
            ind_ext_y=bounds_y,
            ind_ext_x=bounds_x,
            xy_only=xy_only,
            z_only=z_only
        )

    if return_list:
        return image, masks
    else:
        return image, masks[0]


def crop_image_to_size(
        image: GenericImage,
        masks: BaseMask | MaskImage | list[BaseMask],
        crop_size: list[float],
        crop_center: None | list[float] = None,
        in_place: bool = False
) -> tuple[GenericImage, None | BaseMask | MaskImage | list[BaseMask]]:

    image, masks, return_list = standard_image_process_checks(image=image, masks=masks)
    if return_list is None:
        return image, None

    # Set crop_center
    if crop_center is None:
        crop_centers = [mask.get_center_position() for mask in masks]
        if len(crop_centers) == 1:
            crop_center = crop_centers[0]
        else:
            crop_center = np.mean(np.stack(crop_centers, axis=1), axis=1)

    # Convert to list
    crop_center = list(crop_center)
    crop_center = [x if not np.isnan(x) else None for x in crop_center]

    if not in_place:
        image = image.copy()
        masks = [mask.copy() for mask in masks]

    image.crop_to_size(center=crop_center, crop_size=crop_size)
    for mask in masks:
        mask.crop_to_size(center=crop_center, crop_size=crop_size)

    if return_list:
        return image, masks
    else:
        return image, masks[0]
